from __future__ import annotations

import math
from collections.abc import Iterable, Iterator, Sequence
from contextlib import contextmanager

import msgspec

from grim.raylib_api import rd, rl

from .geom import Vec2
from .rand import CrtRand

TERRAIN_TEXTURE_SIZE = 1024
TERRAIN_PATCH_SIZE = 128.0
TERRAIN_PATCH_OVERSCAN = 64.0
TERRAIN_CLEAR_COLOR = rl.Color(63, 56, 25, 255)
TERRAIN_BASE_TINT = rl.Color(178, 178, 178, 230)
TERRAIN_OVERLAY_TINT = rl.Color(178, 178, 178, 230)
TERRAIN_DETAIL_TINT = rl.Color(178, 178, 178, 153)
TERRAIN_DENSITY_BASE = 800
TERRAIN_DENSITY_OVERLAY = 0x23
TERRAIN_DENSITY_DETAIL = 0x0F
TERRAIN_DENSITY_SHIFT = 19
TERRAIN_ROTATION_MAX = 0x13A


_ALPHA_TEST_REF_U8 = 4
_ALPHA_TEST_REF_F32 = float(_ALPHA_TEST_REF_U8) / 255.0

# Grim2D enables alpha test globally with:
#   ALPHATESTENABLE=1, ALPHAFUNC=GREATER, ALPHAREF=4
# See: analysis/ghidra/raw/grim.dll_decompiled.c (FUN_10004520).
#
# raylib does not expose fixed-function alpha test, so we emulate it with a tiny
# discard shader for stamping into the terrain render target.
_ALPHA_TEST_SHADER: rl.Shader | None = None
_ALPHA_TEST_SHADER_TRIED = False

_ALPHA_TEST_VS_330 = r"""
#version 330

in vec3 vertexPosition;
in vec2 vertexTexCoord;
in vec4 vertexColor;

out vec2 fragTexCoord;
out vec4 fragColor;

uniform mat4 mvp;

void main() {
    fragTexCoord = vertexTexCoord;
    fragColor = vertexColor;
    gl_Position = mvp * vec4(vertexPosition, 1.0);
}
"""

_ALPHA_TEST_FS_330 = rf"""
#version 330

in vec2 fragTexCoord;
in vec4 fragColor;

uniform sampler2D texture0;
uniform vec4 colDiffuse;

out vec4 finalColor;

void main() {{
    vec4 texel = texture(texture0, fragTexCoord) * fragColor * colDiffuse;
    if (texel.a <= {_ALPHA_TEST_REF_F32:.10f}) discard;
    finalColor = texel;
}}
"""


def _get_alpha_test_shader() -> rl.Shader | None:
    global _ALPHA_TEST_SHADER, _ALPHA_TEST_SHADER_TRIED
    if _ALPHA_TEST_SHADER_TRIED:
        if _ALPHA_TEST_SHADER is not None and _ALPHA_TEST_SHADER.id > 0:
            return _ALPHA_TEST_SHADER
        return None

    _ALPHA_TEST_SHADER_TRIED = True
    try:
        shader = rl.load_shader_from_memory(_ALPHA_TEST_VS_330, _ALPHA_TEST_FS_330)
    except (RuntimeError, OSError, ValueError):
        _ALPHA_TEST_SHADER = None
        return None

    if shader.id <= 0:
        _ALPHA_TEST_SHADER = None
        return None

    _ALPHA_TEST_SHADER = shader
    return _ALPHA_TEST_SHADER


@contextmanager
def _blend_custom(src_factor: int, dst_factor: int, blend_equation: int) -> Iterator[None]:
    # NOTE: raylib/rlgl tracks custom blend factors as state; some backends only
    # apply them when switching the blend mode. Set factors both before and
    # after BeginBlendMode() to ensure the current draw uses the intended values.
    rl.rl_set_blend_factors(src_factor, dst_factor, blend_equation)
    rl.begin_blend_mode(rl.BlendMode.BLEND_CUSTOM)
    rl.rl_set_blend_factors(src_factor, dst_factor, blend_equation)
    try:
        yield
    finally:
        rl.end_blend_mode()


@contextmanager
def _blend_custom_separate(
    src_rgb: int,
    dst_rgb: int,
    src_alpha: int,
    dst_alpha: int,
    eq_rgb: int,
    eq_alpha: int,
) -> Iterator[None]:
    # NOTE: raylib/rlgl tracks custom blend factors as state; some backends only
    # apply them when switching the blend mode. Set factors both before and
    # after BeginBlendMode() to ensure the current draw uses the intended values.
    rl.rl_set_blend_factors_separate(src_rgb, dst_rgb, src_alpha, dst_alpha, eq_rgb, eq_alpha)
    rl.begin_blend_mode(rl.BlendMode.BLEND_CUSTOM_SEPARATE)
    rl.rl_set_blend_factors_separate(src_rgb, dst_rgb, src_alpha, dst_alpha, eq_rgb, eq_alpha)
    try:
        yield
    finally:
        rl.end_blend_mode()


@contextmanager
def _maybe_alpha_test(enabled: bool) -> Iterator[None]:
    if not enabled:
        yield
        return
    shader = _get_alpha_test_shader()
    if shader is None:
        yield
        return
    rl.begin_shader_mode(shader)
    try:
        yield
    finally:
        rl.end_shader_mode()


class GroundDecal(msgspec.Struct):
    texture: rl.Texture
    src: rl.Rectangle
    pos: Vec2
    width: float
    height: float
    rotation_rad: float = 0.0
    tint: rl.Color = rl.WHITE
    centered: bool = True


class GroundCorpseDecal(msgspec.Struct):
    bodyset_frame: int
    top_left: Vec2
    size: float
    rotation_rad: float
    tint: rl.Color = rl.WHITE


class GroundRenderer(msgspec.Struct):
    texture: rl.Texture
    width: int = TERRAIN_TEXTURE_SIZE
    height: int = TERRAIN_TEXTURE_SIZE
    texture_scale: float = 1.0
    alpha_test: bool = True
    debug_log_stamps: bool = False
    texture_failed: bool = False
    screen_width: float | None = None
    screen_height: float | None = None
    overlay: rl.Texture | None = None
    overlay_detail: rl.Texture | None = None
    terrain_filter: float = 1.0
    render_target: rl.RenderTexture | None = None
    _debug_stamp_log: list[dict[str, object]] = msgspec.field(default_factory=list)
    _render_target_ready: bool = False
    _pending_generate: bool = False
    _pending_generate_seed: int | None = None
    _pending_generate_layers: int = 3
    _render_target_warmup_passes: int = 0
    _fallback_seed: int | None = None
    _fallback_layers: int = 0
    _fallback_patches: list[GroundDecal] = msgspec.field(default_factory=list)
    _fallback_decals: list[GroundDecal] = msgspec.field(default_factory=list)
    _fallback_corpse_decals: list[GroundCorpseDecal] = msgspec.field(default_factory=list)
    _fallback_bodyset_texture: rl.Texture | None = None
    _fallback_corpse_shadow: bool = True

    def debug_clear_stamp_log(self) -> None:
        self._debug_stamp_log.clear()

    def debug_stamp_log(self) -> tuple[dict[str, object], ...]:
        return tuple(self._debug_stamp_log)

    def generation_pending(self) -> bool:
        """True while a scheduled terrain generate is still pending."""
        return self._pending_generate

    def render_target_ready(self) -> bool:
        """True when the terrain render target exists and is ready for drawing."""
        return self.render_target is not None and self._render_target_ready

    def _debug_stamp(self, kind: str, **payload: object) -> None:
        if not self.debug_log_stamps:
            return
        self._debug_stamp_log.append({"kind": kind, **payload})
        if len(self._debug_stamp_log) > 96:
            del self._debug_stamp_log[:32]

    def process_pending(self) -> None:
        # Bound the amount of work per tick. Typical warmup sequence:
        #   1) create RT
        #   2) first fill (may be black/uninitialized on some platforms)
        #   3) warmup retry fill
        steps = 0
        while self._pending_generate and steps < 4:
            steps += 1
            if self.render_target is None:
                self.create_render_target()
                continue

            seed = self._pending_generate_seed
            layers = self._pending_generate_layers
            self._pending_generate = False
            self.generate_partial(seed=seed, layers=layers)
            if self.render_target is None and not self.texture_failed:
                self._pending_generate = True
                continue

            if self._render_target_warmup_passes > 0:
                self._render_target_warmup_passes -= 1
                # On some platforms/drivers the first draw into a new RT can come out as
                # black/uninitialized (all-zero). Retry once before marking it ready.
                self._render_target_ready = False
                self._pending_generate = True
                continue

    def create_render_target(self) -> None:
        if self.texture_failed:
            if self.render_target is not None:
                rl.unload_render_texture(self.render_target)
                self.render_target = None
            self._render_target_ready = False
            return

        scale = self.texture_scale
        if scale < 0.5:
            scale = 0.5
        elif scale > 4.0:
            scale = 4.0
        self.texture_scale = scale

        render_w, render_h = self._render_target_size_for(scale)
        if self._ensure_render_target(render_w, render_h):
            return

        old_scale = scale
        self.texture_scale = scale + scale
        render_w, render_h = self._render_target_size_for(self.texture_scale)
        if self._ensure_render_target(render_w, render_h):
            return

        self.texture_failed = True
        self.texture_scale = old_scale
        if self.render_target is not None:
            rl.unload_render_texture(self.render_target)
            self.render_target = None
        self._render_target_ready = False

    def generate(self, seed: int | None = None) -> None:
        self.generate_partial(seed=seed, layers=3)

    def schedule_generate(self, seed: int | None = None, *, layers: int = 3) -> None:
        self._pending_generate_seed = seed
        self._pending_generate_layers = max(0, min(layers, 3))
        self._pending_generate = True

    def generate_partial(self, seed: int | None = None, *, layers: int) -> None:
        layers = max(0, min(layers, 3))
        # Always keep a deterministic fallback representation of the terrain.
        # When the render target is unavailable (or not ready yet), we can render
        # patches + baked decals directly to the screen, matching the exe's
        # `terrain_texture_failed` path.
        self._fallback_seed = seed
        self._fallback_layers = layers
        self._fallback_patches.clear()
        self._fallback_decals.clear()
        self._fallback_corpse_decals.clear()
        self._fallback_bodyset_texture = None
        self._fallback_corpse_shadow = True

        rng_fallback = CrtRand(seed)
        if layers >= 1:
            self._scatter_texture_fallback(self.texture, TERRAIN_BASE_TINT, rng_fallback, TERRAIN_DENSITY_BASE)
        if layers >= 2 and self.overlay is not None:
            self._scatter_texture_fallback(self.overlay, TERRAIN_OVERLAY_TINT, rng_fallback, TERRAIN_DENSITY_OVERLAY)
        if layers >= 3:
            # Original uses base texture for detail pass, not overlay.
            self._scatter_texture_fallback(self.texture, TERRAIN_DETAIL_TINT, rng_fallback, TERRAIN_DENSITY_DETAIL)

        self.create_render_target()
        if self.render_target is None:
            return
        rng = CrtRand(seed)
        self._set_stamp_filters(point=True)
        rl.begin_texture_mode(self.render_target)
        rl.clear_background(TERRAIN_CLEAR_COLOR)
        # Keep the ground RT alpha at 1.0 like the original exe (which typically uses
        # an XRGB render target). We still alpha-blend RGB, but preserve destination A.
        with _blend_custom_separate(
            rd.RL_SRC_ALPHA,
            rd.RL_ONE_MINUS_SRC_ALPHA,
            rd.RL_ZERO,
            rd.RL_ONE,
            rd.RL_FUNC_ADD,
            rd.RL_FUNC_ADD,
        ):
            if layers >= 1:
                self._scatter_texture(self.texture, TERRAIN_BASE_TINT, rng, TERRAIN_DENSITY_BASE)
            if layers >= 2 and self.overlay is not None:
                self._scatter_texture(self.overlay, TERRAIN_OVERLAY_TINT, rng, TERRAIN_DENSITY_OVERLAY)
            if layers >= 3:
                # Original uses base texture for detail pass, not overlay
                self._scatter_texture(self.texture, TERRAIN_DETAIL_TINT, rng, TERRAIN_DENSITY_DETAIL)
        rl.end_texture_mode()
        self._set_stamp_filters(point=False)
        self._render_target_ready = True

    def bake_decals(self, decals: Sequence[GroundDecal]) -> bool:
        if not decals:
            return False

        self.create_render_target()
        if self.render_target is None:
            self._fallback_decals.extend(decals)
            return True

        if self.debug_log_stamps:
            head = decals[0]
            self._debug_stamp(
                "bake_decals",
                count=len(decals),
                pos0=head.pos.to_dict(),
                rot0=float(head.rotation_rad),
            )

        inv_scale = 1.0 / self._normalized_texture_scale()
        rl.begin_texture_mode(self.render_target)
        with _blend_custom_separate(
            rd.RL_SRC_ALPHA,
            rd.RL_ONE_MINUS_SRC_ALPHA,
            rd.RL_ZERO,
            rd.RL_ONE,
            rd.RL_FUNC_ADD,
            rd.RL_FUNC_ADD,
        ):
            for decal in decals:
                w = decal.width
                h = decal.height
                if decal.centered:
                    pivot_x = decal.pos.x
                    pivot_y = decal.pos.y
                else:
                    pivot_x = decal.pos.x + w * 0.5
                    pivot_y = decal.pos.y + h * 0.5
                pivot_x *= inv_scale
                pivot_y *= inv_scale
                w *= inv_scale
                h *= inv_scale
                dst = rl.Rectangle(pivot_x, pivot_y, w, h)
                origin = rl.Vector2(w * 0.5, h * 0.5)
                rl.draw_texture_pro(
                    decal.texture,
                    decal.src,
                    dst,
                    origin,
                    math.degrees(decal.rotation_rad),
                    decal.tint,
                )
        rl.end_texture_mode()

        self._render_target_ready = True
        return True

    def bake_corpse_decals(
        self,
        bodyset_texture: rl.Texture,
        decals: Sequence[GroundCorpseDecal],
        *,
        shadow: bool = True,
    ) -> bool:
        if not decals:
            return False

        self.create_render_target()
        if self.render_target is None:
            self._fallback_bodyset_texture = bodyset_texture
            self._fallback_corpse_shadow = shadow
            self._fallback_corpse_decals.extend(decals)
            return True

        if self.debug_log_stamps:
            head = decals[0]
            self._debug_stamp(
                "bake_corpse_decals",
                shadow=shadow,
                count=len(decals),
                frame0=int(head.bodyset_frame),
                top_left0=head.top_left.to_dict(),
                size0=float(head.size),
                rot0=float(head.rotation_rad),
            )

        scale = self._normalized_texture_scale()
        inv_scale = 1.0 / scale
        offset = 2.0 * scale / float(self.width)
        rl.begin_texture_mode(self.render_target)
        with _maybe_alpha_test(self.alpha_test):
            if shadow:
                if self.debug_log_stamps:
                    self._debug_stamp("corpse_shadow_pass", draws=len(decals))
                self._draw_corpse_shadow_pass(bodyset_texture, decals, inv_scale, offset)
            if self.debug_log_stamps:
                self._debug_stamp("corpse_color_pass", draws=len(decals))
            self._draw_corpse_color_pass(bodyset_texture, decals, inv_scale, offset)
        rl.end_texture_mode()

        self._render_target_ready = True
        return True

    def _draw_fallback(
        self,
        camera: Vec2,
        *,
        out_w: float,
        out_h: float,
        screen_w: float,
        screen_h: float,
    ) -> None:
        rl.draw_rectangle(0, 0, int(out_w + 0.5), int(out_h + 0.5), TERRAIN_CLEAR_COLOR)
        if screen_w <= 0.0 or screen_h <= 0.0:
            return

        view_scale = Vec2(out_w / screen_w, out_h / screen_h)

        view_x0 = -camera.x
        view_y0 = -camera.y
        view_x1 = view_x0 + screen_w
        view_y1 = view_y0 + screen_h

        def draw_decal(decal: GroundDecal) -> None:
            texture = decal.texture
            if texture.id <= 0:
                return
            w = float(decal.width)
            h = float(decal.height)
            if w <= 0.0 or h <= 0.0:
                return

            pivot = decal.pos
            if not decal.centered:
                pivot += Vec2(w * 0.5, h * 0.5)

            if pivot.x + w * 0.5 < view_x0 or pivot.x - w * 0.5 > view_x1:
                return
            if pivot.y + h * 0.5 < view_y0 or pivot.y - h * 0.5 > view_y1:
                return

            screen_pivot = (pivot + camera).mul_components(view_scale)
            scaled_size = Vec2(w, h).mul_components(view_scale)
            sw = scaled_size.x
            sh = scaled_size.y
            dst = rl.Rectangle(screen_pivot.x, screen_pivot.y, float(sw), float(sh))
            origin = rl.Vector2(float(sw) * 0.5, float(sh) * 0.5)
            rl.draw_texture_pro(
                texture,
                decal.src,
                dst,
                origin,
                math.degrees(float(decal.rotation_rad)),
                decal.tint,
            )

        with _blend_custom_separate(
            rd.RL_SRC_ALPHA,
            rd.RL_ONE_MINUS_SRC_ALPHA,
            rd.RL_ZERO,
            rd.RL_ONE,
            rd.RL_FUNC_ADD,
            rd.RL_FUNC_ADD,
        ):
            for patch in self._fallback_patches:
                draw_decal(patch)
            for decal in self._fallback_decals:
                draw_decal(decal)

        bodyset_texture = self._fallback_bodyset_texture
        if bodyset_texture is None or not self._fallback_corpse_decals:
            return

        def draw_corpse(size: float, pivot: Vec2, rotation_deg: float, tint: rl.Color, src: rl.Rectangle) -> None:
            if pivot.x + size * 0.5 < view_x0 or pivot.x - size * 0.5 > view_x1:
                return
            if pivot.y + size * 0.5 < view_y0 or pivot.y - size * 0.5 > view_y1:
                return
            screen_pivot = (pivot + camera).mul_components(view_scale)
            scaled_size = Vec2(size, size).mul_components(view_scale)
            sw = scaled_size.x
            sh = scaled_size.y
            dst = rl.Rectangle(screen_pivot.x, screen_pivot.y, float(sw), float(sh))
            origin = rl.Vector2(float(sw) * 0.5, float(sh) * 0.5)
            rl.draw_texture_pro(bodyset_texture, src, dst, origin, rotation_deg, tint)

        with _maybe_alpha_test(self.alpha_test):
            if self._fallback_corpse_shadow:
                with _blend_custom_separate(
                    rd.RL_ZERO,
                    rd.RL_ONE_MINUS_SRC_ALPHA,
                    rd.RL_ZERO,
                    rd.RL_ONE,
                    rd.RL_FUNC_ADD,
                    rd.RL_FUNC_ADD,
                ):
                    for decal in self._fallback_corpse_decals:
                        src = self._corpse_src(bodyset_texture, decal.bodyset_frame)
                        size = float(decal.size) * 1.064
                        pivot = decal.top_left + Vec2(size * 0.5 - 0.5, size * 0.5 - 0.5)
                        tint = rl.Color(decal.tint.r, decal.tint.g, decal.tint.b, int(decal.tint.a * 0.5))
                        rotation_deg = math.degrees(float(decal.rotation_rad) - (math.pi * 0.5))
                        draw_corpse(size, pivot, rotation_deg, tint, src)

            with _blend_custom_separate(
                rd.RL_SRC_ALPHA,
                rd.RL_ONE_MINUS_SRC_ALPHA,
                rd.RL_ZERO,
                rd.RL_ONE,
                rd.RL_FUNC_ADD,
                rd.RL_FUNC_ADD,
            ):
                for decal in self._fallback_corpse_decals:
                    src = self._corpse_src(bodyset_texture, decal.bodyset_frame)
                    size = float(decal.size)
                    pivot = decal.top_left + Vec2(size * 0.5, size * 0.5)
                    rotation_deg = math.degrees(float(decal.rotation_rad) - (math.pi * 0.5))
                    draw_corpse(size, pivot, rotation_deg, decal.tint, src)

    def draw(
        self,
        camera: Vec2,
        *,
        screen_w: float | None = None,
        screen_h: float | None = None,
        out_w: float | None = None,
        out_h: float | None = None,
    ) -> None:
        if out_w is None:
            out_w = float(rl.get_screen_width())
        else:
            out_w = float(out_w)
        if out_h is None:
            out_h = float(rl.get_screen_height())
        else:
            out_h = float(out_h)
        if out_w <= 0.0:
            out_w = float(rl.get_screen_width())
        if out_h <= 0.0:
            out_h = float(rl.get_screen_height())
        if screen_w is None:
            # Prefer live output dimensions by default. Cached config-sized
            # values can lag during gameplay/menu handoffs.
            if out_w > 0.0:
                screen_w = out_w
            else:
                screen_w = float(self.screen_width or out_w)
        if screen_h is None:
            if out_h > 0.0:
                screen_h = out_h
            else:
                screen_h = float(self.screen_height or out_h)
        if screen_w <= 0.0:
            screen_w = out_w if out_w > 0.0 else float(self.screen_width or 1.0)
        if screen_h <= 0.0:
            screen_h = out_h if out_h > 0.0 else float(self.screen_height or 1.0)
        screen_w, screen_h = self._fit_view_window(screen_w, screen_h)
        cam = self._clamp_camera(camera, screen_w, screen_h)

        if self.render_target is None or not self._render_target_ready:
            self._draw_fallback(cam, out_w=out_w, out_h=out_h, screen_w=float(screen_w), screen_h=float(screen_h))
            return

        target = self.render_target
        u0 = -cam.x / float(self.width)
        v0 = -cam.y / float(self.height)
        u1 = u0 + screen_w / float(self.width)
        v1 = v0 + screen_h / float(self.height)
        src_x = u0 * float(target.texture.width)
        # Render textures are vertically flipped in raylib, so adjust the source
        # rectangle to sample the correct world-space slice before flipping.
        src_y = (1.0 - v1) * float(target.texture.height)
        src_w = (u1 - u0) * float(target.texture.width)
        src_h = (v1 - v0) * float(target.texture.height)
        src = rl.Rectangle(src_x, src_y, src_w, -src_h)
        dst = rl.Rectangle(0.0, 0.0, out_w, out_h)
        if self.terrain_filter == 2.0:
            rl.set_texture_filter(target.texture, rl.TextureFilter.TEXTURE_FILTER_POINT)
        # Disable alpha blending when drawing terrain to screen - the render target's
        # alpha channel may be < 1.0 after stamp blending, but terrain should be opaque.
        with _blend_custom(rd.RL_ONE, rd.RL_ZERO, rd.RL_FUNC_ADD):
            rl.draw_texture_pro(target.texture, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.WHITE)
        if self.terrain_filter == 2.0:
            rl.set_texture_filter(target.texture, rl.TextureFilter.TEXTURE_FILTER_BILINEAR)

    def _fit_view_window(self, screen_w: float, screen_h: float) -> tuple[float, float]:
        """
        Convert output dimensions into a world-space camera window.

        Keep a uniform pixel scale and never request a camera window larger than
        the terrain dimensions. This avoids non-uniform stretch on widescreen
        outputs where only one axis exceeds world size.
        """

        world_w = float(self.width)
        world_h = float(self.height)
        if world_w <= 0.0 or world_h <= 0.0:
            return max(1.0, float(screen_w)), max(1.0, float(screen_h))

        out_w = max(1.0, float(screen_w))
        out_h = max(1.0, float(screen_h))
        scale = max(out_w / world_w, out_h / world_h, 1.0)
        view_w = min(world_w, out_w / scale)
        view_h = min(world_h, out_h / scale)
        return view_w, view_h

    def _scatter_texture(
        self,
        texture: rl.Texture,
        tint: rl.Color,
        rng: CrtRand,
        density: int,
    ) -> None:
        area = self.width * self.height
        count = (area * density) >> TERRAIN_DENSITY_SHIFT
        if count <= 0:
            return
        inv_scale = 1.0 / self._normalized_texture_scale()
        size = TERRAIN_PATCH_SIZE * inv_scale
        src = rl.Rectangle(0.0, 0.0, float(texture.width), float(texture.height))
        origin = rl.Vector2(size * 0.5, size * 0.5)
        span_w = self.width + int(TERRAIN_PATCH_OVERSCAN * 2)
        # The original exe uses `terrain_texture_width` for both axes. Terrain is
        # square (1024x1024) so this is equivalent, but keep it for parity.
        span_h = span_w
        for _ in range(count):
            angle = ((rng.rand() % TERRAIN_ROTATION_MAX) * 0.01) % math.tau
            # IMPORTANT: The exe consumes RNG as rotation, then Y, then X.
            y = ((rng.rand() % span_h) - TERRAIN_PATCH_OVERSCAN) * inv_scale
            x = ((rng.rand() % span_w) - TERRAIN_PATCH_OVERSCAN) * inv_scale
            # raylib's DrawTexturePro positions the quad by the *origin point*,
            # while the original engine uses x/y as the quad top-left.
            dst = rl.Rectangle(float(x + size * 0.5), float(y + size * 0.5), size, size)
            rl.draw_texture_pro(texture, src, dst, origin, math.degrees(angle), tint)

    def _scatter_texture_fallback(
        self,
        texture: rl.Texture,
        tint: rl.Color,
        rng: CrtRand,
        density: int,
    ) -> None:
        """Record terrain patch draws for the render-target fallback path."""
        area = self.width * self.height
        count = (area * density) >> TERRAIN_DENSITY_SHIFT
        if count <= 0:
            return

        size = float(TERRAIN_PATCH_SIZE)
        src = rl.Rectangle(0.0, 0.0, float(texture.width), float(texture.height))
        span_w = self.width + int(TERRAIN_PATCH_OVERSCAN * 2)
        # The original exe uses `terrain_texture_width` for both axes. Terrain is
        # square (1024x1024) so this is equivalent, but keep it for parity.
        span_h = span_w

        for _ in range(count):
            angle = ((rng.rand() % TERRAIN_ROTATION_MAX) * 0.01) % math.tau
            # IMPORTANT: The exe consumes RNG as rotation, then Y, then X.
            y = float((rng.rand() % span_h) - TERRAIN_PATCH_OVERSCAN)
            x = float((rng.rand() % span_w) - TERRAIN_PATCH_OVERSCAN)
            self._fallback_patches.append(
                GroundDecal(
                    texture=texture,
                    src=src,
                    pos=Vec2(float(x + size * 0.5), float(y + size * 0.5)),
                    width=size,
                    height=size,
                    rotation_rad=float(angle),
                    tint=tint,
                    centered=True,
                ),
            )

    def _clamp_camera(self, camera: Vec2, screen_w: float, screen_h: float) -> Vec2:
        cam_x = camera.x
        cam_y = camera.y
        if cam_x > -1.0:
            cam_x = -1.0
        if cam_y > -1.0:
            cam_y = -1.0
        min_x = screen_w - float(self.width)
        min_y = screen_h - float(self.height)
        if cam_x < min_x:
            cam_x = min_x
        if cam_y < min_y:
            cam_y = min_y
        return Vec2(cam_x, cam_y)

    def _ensure_render_target(self, render_w: int, render_h: int) -> bool:
        if self.render_target is not None:
            if self.render_target.texture.width == render_w and self.render_target.texture.height == render_h:
                return True
            rl.unload_render_texture(self.render_target)
            self.render_target = None
            self._render_target_ready = False

        try:
            candidate = rl.load_render_texture(render_w, render_h)
        except (RuntimeError, OSError, ValueError):
            return False

        if candidate.id == 0 or not rl.is_render_texture_valid(candidate):
            if candidate.id:
                rl.unload_render_texture(candidate)
            return False
        if (
            candidate.texture.width <= 0
            or candidate.texture.height <= 0
        ):
            rl.unload_render_texture(candidate)
            return False

        self.render_target = candidate
        self._render_target_ready = False
        self._render_target_warmup_passes = 1
        rl.set_texture_filter(self.render_target.texture, rl.TextureFilter.TEXTURE_FILTER_BILINEAR)
        rl.set_texture_wrap(self.render_target.texture, rl.TextureWrap.TEXTURE_WRAP_CLAMP)
        return True

    def _render_target_size_for(self, scale: float) -> tuple[int, int]:
        pixel_scale = 1.0
        screen_w = int(rl.get_screen_width())
        screen_h = int(rl.get_screen_height())
        render_w = int(rl.get_render_width())
        render_h = int(rl.get_render_height())
        if render_w == screen_w * 2 and render_h == screen_h * 2:
            pixel_scale = 2.0
        render_w = max(1, int((self.width * pixel_scale) / scale))
        render_h = max(1, int((self.height * pixel_scale) / scale))
        return render_w, render_h

    def _normalized_texture_scale(self) -> float:
        scale = self.texture_scale
        if scale < 0.5:
            scale = 0.5
        screen_w = int(rl.get_screen_width())
        screen_h = int(rl.get_screen_height())
        render_w = int(rl.get_render_width())
        render_h = int(rl.get_render_height())
        if render_w == screen_w * 2 and render_h == screen_h * 2:
            scale *= 0.5
        return scale

    def _set_stamp_filters(self, *, point: bool) -> None:
        self._set_texture_filters(
            (self.texture, self.overlay, self.overlay_detail),
            point=point,
        )

    @staticmethod
    def _set_texture_filters(textures: Iterable[rl.Texture | None], *, point: bool) -> None:
        mode = rl.TextureFilter.TEXTURE_FILTER_POINT if point else rl.TextureFilter.TEXTURE_FILTER_BILINEAR
        for texture in textures:
            if texture is None:
                continue
            if texture.id <= 0:
                continue
            rl.set_texture_filter(texture, mode)

    def _corpse_src(self, bodyset_texture: rl.Texture, frame: int) -> rl.Rectangle:
        frame = int(frame) & 0xF
        cell_w = float(bodyset_texture.width) * 0.25
        cell_h = float(bodyset_texture.height) * 0.25
        col = frame & 3
        row = frame >> 2
        return rl.Rectangle(cell_w * float(col), cell_h * float(row), cell_w, cell_h)

    def _draw_corpse_shadow_pass(
        self,
        bodyset_texture: rl.Texture,
        decals: Sequence[GroundCorpseDecal],
        inv_scale: float,
        offset: float,
    ) -> None:
        with _blend_custom_separate(
            rd.RL_ZERO,
            rd.RL_ONE_MINUS_SRC_ALPHA,
            rd.RL_ZERO,
            rd.RL_ONE,
            rd.RL_FUNC_ADD,
            rd.RL_FUNC_ADD,
        ):
            for decal in decals:
                src = self._corpse_src(bodyset_texture, decal.bodyset_frame)
                size = decal.size * inv_scale * 1.064
                x = (decal.top_left.x - 0.5) * inv_scale - offset
                y = (decal.top_left.y - 0.5) * inv_scale - offset
                dst = rl.Rectangle(x + size * 0.5, y + size * 0.5, size, size)
                origin = rl.Vector2(size * 0.5, size * 0.5)
                tint = rl.Color(
                    decal.tint.r,
                    decal.tint.g,
                    decal.tint.b,
                    int(decal.tint.a * 0.5),
                )
                rl.draw_texture_pro(
                    bodyset_texture,
                    src,
                    dst,
                    origin,
                    math.degrees(decal.rotation_rad - (math.pi * 0.5)),
                    tint,
                )

    def _draw_corpse_color_pass(
        self,
        bodyset_texture: rl.Texture,
        decals: Sequence[GroundCorpseDecal],
        inv_scale: float,
        offset: float,
    ) -> None:
        with _blend_custom_separate(
            rd.RL_SRC_ALPHA,
            rd.RL_ONE_MINUS_SRC_ALPHA,
            rd.RL_ZERO,
            rd.RL_ONE,
            rd.RL_FUNC_ADD,
            rd.RL_FUNC_ADD,
        ):
            for decal in decals:
                src = self._corpse_src(bodyset_texture, decal.bodyset_frame)
                size = decal.size * inv_scale
                x = decal.top_left.x * inv_scale - offset
                y = decal.top_left.y * inv_scale - offset
                dst = rl.Rectangle(x + size * 0.5, y + size * 0.5, size, size)
                origin = rl.Vector2(size * 0.5, size * 0.5)
                rl.draw_texture_pro(
                    bodyset_texture,
                    src,
                    dst,
                    origin,
                    math.degrees(decal.rotation_rad - (math.pi * 0.5)),
                    decal.tint,
                )



from __future__ import annotations

import msgspec


class SlotState(msgspec.Struct, forbid_unknown_fields=True):
    slot_index: int = -1
    connected: bool = False
    ready: bool = False
    is_host: bool = False
    peer_name: str = ""


class PacketHeader(msgspec.Struct, forbid_unknown_fields=True):
    seq: int = 0
    ack: int = 0
    reliable: bool = False


__all__ = ["PacketHeader", "SlotState"]



from __future__ import annotations

import msgspec

from .spawn_ids import CreatureFlags, CreatureTypeId, SpawnId, Tint


class SpawnTemplate(msgspec.Struct, frozen=True):
    spawn_id: SpawnId
    type_id: CreatureTypeId | None
    flags: CreatureFlags | None
    creature: str | None
    anim_note: str | None
    tint: Tint | None = None
    size: float | None = None
    move_speed: float | None = None


TYPE_ID_TO_NAME = {type_id.value: type_id.name.lower() for type_id in CreatureTypeId}

# For many template ids, tint/size/move_speed are randomized or derived from other fields.
# We only fill them in when the game uses fixed constants (and keep the rest as `None`).
SPAWN_TEMPLATES = [
    SpawnTemplate(
        spawn_id=SpawnId.ZOMBIE_BOSS_SPAWNER_00,
        type_id=CreatureTypeId.ZOMBIE,
        flags=CreatureFlags.ANIM_PING_PONG | CreatureFlags.ANIM_LONG_STRIP,
        creature="zombie",
        anim_note="long strip (0x40 overrides 0x4)",
        tint=(0.6, 0.6, 1.0, 0.8),
        size=64.0,
        move_speed=1.3,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP2_SPLITTER_01,
        type_id=CreatureTypeId.SPIDER_SP2,
        flags=CreatureFlags.SPLIT_ON_DEATH,
        creature="spider_sp2",
        anim_note=None,
        tint=(0.8, 0.7, 0.4, 1.0),
        size=80.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_RANDOM_03,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.LIZARD_RANDOM_04,
        type_id=CreatureTypeId.LIZARD,
        flags=None,
        creature="lizard",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP2_RANDOM_05,
        type_id=CreatureTypeId.SPIDER_SP2,
        flags=None,
        creature="spider_sp2",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_RANDOM_06,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_1D_FAST_07,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(1.0, 1.0, 1.0, 1.0),
        size=50.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_1D_SLOW_08,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(1.0, 1.0, 1.0, 1.0),
        size=50.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_1D_LIMITED_09,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(1.0, 1.0, 1.0, 1.0),
        size=40.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_32_SLOW_0A,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(0.8, 0.7, 0.4, 1.0),
        size=55.0,
        move_speed=1.5,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_3C_SLOW_0B,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(0.9, 0.1, 0.1, 1.0),
        size=65.0,
        move_speed=1.5,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_FAST_0C,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(0.9, 0.8, 0.4, 1.0),
        size=32.0,
        move_speed=2.8,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_31_SLOW_0D,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(0.9, 0.8, 0.4, 1.0),
        size=32.0,
        move_speed=1.3,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_RING_24_0E,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(0.9, 0.8, 0.4, 1.0),
        size=32.0,
        move_speed=2.8,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_BROWN_TRANSPARENT_0F,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.665, 0.385, 0.259, 0.56),
        size=50.0,
        move_speed=2.9,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_SPAWNER_CHILD_32_FAST_10,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.ANIM_PING_PONG,
        creature="alien",
        anim_note="short strip (ping-pong)",
        tint=(0.9, 0.8, 0.4, 1.0),
        size=32.0,
        move_speed=2.8,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_CHAIN_LIZARD_4_11,
        type_id=CreatureTypeId.LIZARD,
        flags=None,
        creature="lizard",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_RING_ALIEN_8_12,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_CHAIN_ALIEN_10_13,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_GRID_ALIEN_GREEN_14,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_GRID_ALIEN_WHITE_15,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_GRID_LIZARD_WHITE_16,
        type_id=CreatureTypeId.LIZARD,
        flags=None,
        creature="lizard",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_GRID_SPIDER_SP1_WHITE_17,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_GRID_ALIEN_BRONZE_18,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.FORMATION_RING_ALIEN_5_19,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.AI1_ALIEN_BLUE_TINT_1A,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.AI1_SPIDER_SP1_BLUE_TINT_1B,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.AI1_LIZARD_BLUE_TINT_1C,
        type_id=CreatureTypeId.LIZARD,
        flags=None,
        creature="lizard",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_RANDOM_1D,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_RANDOM_1E,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_RANDOM_1F,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_RANDOM_GREEN_20,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_PURPLE_GHOST_21,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_GREEN_GHOST_22,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_GREEN_GHOST_SMALL_23,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_GREEN_24,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.1, 0.7, 0.11, 1.0),
        size=50.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_GREEN_SMALL_25,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.1, 0.8, 0.11, 1.0),
        size=30.0,
        move_speed=2.5,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_PALE_GREEN_26,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.6, 0.8, 0.6, 1.0),
        size=45.0,
        move_speed=2.2,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_WEAPON_BONUS_27,
        type_id=CreatureTypeId.ALIEN,
        flags=CreatureFlags.BONUS_ON_DEATH,
        creature="alien",
        anim_note="bonus_id=WEAPON (3), duration_override=5 (packed in link_index)",
        tint=(1.0, 0.8, 0.1, 1.0),
        size=45.0,
        move_speed=2.1,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_PURPLE_28,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.7, 0.1, 0.51, 1.0),
        size=55.0,
        move_speed=1.7,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_GREY_BRUTE_29,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.8, 0.8, 0.8, 1.0),
        size=70.0,
        move_speed=2.5,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_GREY_FAST_2A,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.3, 0.3, 0.3, 1.0),
        size=60.0,
        move_speed=3.1,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_RED_FAST_2B,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(1.0, 0.3, 0.3, 1.0),
        size=35.0,
        move_speed=3.6,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_RED_BOSS_2C,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.85, 0.2, 0.2, 1.0),
        size=80.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_CONST_CYAN_AI2_2D,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.0, 0.9, 0.8, 1.0),
        size=38.0,
        move_speed=3.1,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.LIZARD_RANDOM_2E,
        type_id=CreatureTypeId.LIZARD,
        flags=None,
        creature="lizard",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.LIZARD_CONST_GREY_2F,
        type_id=CreatureTypeId.LIZARD,
        flags=None,
        creature="lizard",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.LIZARD_CONST_YELLOW_BOSS_30,
        type_id=CreatureTypeId.LIZARD,
        flags=None,
        creature="lizard",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.LIZARD_RANDOM_31,
        type_id=CreatureTypeId.LIZARD,
        flags=None,
        creature="lizard",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_RANDOM_32,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_RANDOM_RED_33,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_RANDOM_GREEN_34,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP2_RANDOM_35,
        type_id=CreatureTypeId.SPIDER_SP2,
        flags=None,
        creature="spider_sp2",
        anim_note=None,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ALIEN_AI7_ORBITER_36,
        type_id=CreatureTypeId.ALIEN,
        flags=None,
        creature="alien",
        anim_note=None,
        tint=(0.65, None, 0.95, 1.0),
        size=50.0,
        move_speed=1.8,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP2_RANGED_VARIANT_37,
        type_id=CreatureTypeId.SPIDER_SP2,
        flags=CreatureFlags.RANGED_ATTACK_VARIANT,
        creature="spider_sp2",
        anim_note=None,
        tint=(1.0, 0.75, 0.1, 1.0),
        move_speed=3.2,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_AI7_TIMER_38,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=CreatureFlags.AI7_LINK_TIMER,
        creature="spider_sp1",
        anim_note=None,
        tint=(1.0, 0.75, 0.1, 1.0),
        move_speed=4.8,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_AI7_TIMER_WEAK_39,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=CreatureFlags.AI7_LINK_TIMER,
        creature="spider_sp1",
        anim_note=None,
        tint=(0.8, 0.65, 0.1, 1.0),
        move_speed=4.8,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_CONST_SHOCK_BOSS_3A,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=CreatureFlags.RANGED_ATTACK_SHOCK,
        creature="spider_sp1",
        anim_note="projectile_type=9",
        tint=(1.0, 1.0, 1.0, 1.0),
        size=64.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_CONST_RED_BOSS_3B,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
        tint=(0.9, 0.0, 0.0, 1.0),
        size=70.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_CONST_RANGED_VARIANT_3C,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=CreatureFlags.RANGED_ATTACK_VARIANT,
        creature="spider_sp1",
        anim_note="projectile_type=26 (packed in orbit_radius)",
        tint=(0.9, 0.1, 0.1, 1.0),
        size=40.0,
        move_speed=2.0,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_RANDOM_3D,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
        tint=(None, None, None, 1.0),
        move_speed=2.6,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_CONST_WHITE_FAST_3E,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
        tint=(1.0, 1.0, 1.0, 1.0),
        size=64.0,
        move_speed=2.8,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_CONST_BROWN_SMALL_3F,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
        tint=(0.7, 0.4, 0.1, 1.0),
        size=35.0,
        move_speed=2.3,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.SPIDER_SP1_CONST_BLUE_40,
        type_id=CreatureTypeId.SPIDER_SP1,
        flags=None,
        creature="spider_sp1",
        anim_note=None,
        tint=(0.5, 0.6, 0.9, 1.0),
        size=45.0,
        move_speed=2.2,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ZOMBIE_RANDOM_41,
        type_id=CreatureTypeId.ZOMBIE,
        flags=None,
        creature="zombie",
        anim_note=None,
        tint=(None, None, None, 1.0),
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ZOMBIE_CONST_GREY_42,
        type_id=CreatureTypeId.ZOMBIE,
        flags=None,
        creature="zombie",
        anim_note=None,
        tint=(0.9, 0.9, 0.9, 1.0),
        size=45.0,
        move_speed=1.7,
    ),
    SpawnTemplate(
        spawn_id=SpawnId.ZOMBIE_CONST_GREEN_BRUTE_43,
        type_id=CreatureTypeId.ZOMBIE,
        flags=None,
        creature="zombie",
        anim_note=None,
        tint=(0.2, 0.6, 0.1, 1.0),
        size=70.0,
        move_speed=2.1,
    ),
]

SPAWN_ID_TO_TEMPLATE = {entry.spawn_id: entry for entry in SPAWN_TEMPLATES}




from __future__ import annotations

import random
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import cast

from grim.assets import PaqTextureCache
from grim.audio import AudioState
from grim.config import CrimsonConfig
from grim.geom import Vec2

from ..game_modes import GameMode
from ..render.frame import RenderFrame
from ..render.rtx.mode import RtxRenderMode
from ..render.world.renderer import WorldRenderer, WorldRenderHost
from ..sim.batch_apply import apply_presentation_outputs, apply_sim_metadata_batch
from ..sim.clock import FixedStepClock
from ..sim.frame_pump import advance_tick_runner_frame
from ..sim.input import PlayerInput
from ..sim.input_providers import FrameContext, LocalInputProvider
from ..sim.presentation_reactions import (
    PostApplyReaction,
    apply_post_apply_reaction,
    build_post_apply_reaction,
)
from ..sim.sessions import DeterministicSession
from ..sim.tick_runner import TickBatchResult, TickRunner, TickRunnerConfig
from .audio_bridge import AudioBridge
from .presentation import PresentationLayer
from .render_resources import RenderResources
from .sim_world_state import SimWorldState
from .terrain_runtime import TerrainRuntime

WorldTickInputBuilder = Callable[[FrameContext], Sequence[PlayerInput]]


class WorldRuntime:
    """Composition container owning the 4 world components and shared lifecycle methods.

    Satisfies both ``WorldHost`` (sim/tick harness) and ``WorldRenderHost``
    (renderer) protocols, eliminating unsafe ``cast()`` patterns.
    """

    def __init__(
        self,
        *,
        assets_dir: Path,
        world_size: float = 1024.0,
        demo_mode_active: bool = False,
        quest_fail_retry_count: int = 0,
        hardcore: bool = False,
        preserve_bugs: bool = False,
        texture_cache: PaqTextureCache | None = None,
        config: CrimsonConfig | None = None,
        audio: AudioState | None = None,
        audio_rng: random.Random | None = None,
        rtx_mode: RtxRenderMode = RtxRenderMode.CLASSIC,
    ) -> None:
        self.assets_dir = Path(assets_dir)
        self.world_size = float(world_size)
        self.demo_mode_active = bool(demo_mode_active)
        self.quest_fail_retry_count = int(quest_fail_retry_count)
        self.hardcore = bool(hardcore)
        self.preserve_bugs = bool(preserve_bugs)
        self.texture_cache = texture_cache
        self.config = config
        self.audio = audio
        self.audio_rng = audio_rng
        self.rtx_mode = rtx_mode

        self.sim_world = SimWorldState(
            world_size=float(self.world_size),
            demo_mode_active=bool(self.demo_mode_active),
            hardcore=bool(self.hardcore),
            quest_fail_retry_count=int(self.quest_fail_retry_count),
            preserve_bugs=bool(self.preserve_bugs),
        )
        render_resources = RenderResources(
            assets_dir=self.assets_dir,
            world_size=float(self.world_size),
            texture_cache=self.texture_cache,
            config=self.config,
        )
        self.presentation = PresentationLayer(
            render_resources=render_resources,
            audio_bridge=AudioBridge(
                demo_mode_active=bool(self.demo_mode_active),
                reflex_boost_timer_source=lambda: float(self.sim_world.state.bonuses.reflex_boost),
                audio=self.audio,
                audio_rng=self.audio_rng,
            ),
            terrain_runtime=TerrainRuntime(
                world_size=float(self.world_size),
                render_resources=render_resources,
            ),
        )

        self.camera = Vec2(-1.0, -1.0)
        self.lan_player_rings_enabled = False
        self.lan_local_aim_indicators_only = False
        self.lan_local_player_slot_index = 0

        self._sync_world_size_ownership()
        self.sync_audio_bridge_state()
        self.renderer = WorldRenderer(cast(WorldRenderHost, self))

    # ------------------------------------------------------------------
    # Forwarding properties (satisfy WorldHost / WorldRenderHost protocols)
    # ------------------------------------------------------------------

    @property
    def render_resources(self) -> RenderResources:
        return self.presentation.render_resources

    @property
    def audio_bridge(self) -> AudioBridge:
        return self.presentation.audio_bridge

    @property
    def terrain_runtime(self) -> TerrainRuntime:
        return self.presentation.terrain_runtime

    # ------------------------------------------------------------------
    # Shared lifecycle methods (extracted from 4 identical implementations)
    # ------------------------------------------------------------------

    def sync_world_size(self) -> None:
        self._sync_world_size_ownership()

    def _sync_world_size_ownership(self) -> None:
        world_size = float(self.world_size)
        self.sim_world.world_size = world_size
        self.render_resources.world_size = world_size
        self.terrain_runtime.world_size = world_size
        ground = self.render_resources.ground
        if ground is not None:
            side = max(0, int(world_size))
            ground.width = side
            ground.height = side

    def reset(
        self,
        *,
        seed: int = 0xBEEF,
        player_count: int = 1,
        spawn_pos: Vec2 | None = None,
    ) -> None:
        self._sync_world_size_ownership()
        self.sim_world.demo_mode_active = bool(self.demo_mode_active)
        self.sim_world.hardcore = bool(self.hardcore)
        self.sim_world.quest_fail_retry_count = int(self.quest_fail_retry_count)
        self.sim_world.preserve_bugs = bool(self.preserve_bugs)
        self.sim_world.reset(
            seed=int(seed),
            player_count=int(player_count),
            spawn_pos=spawn_pos,
        )
        self.render_resources.fx_queue.clear()
        self.render_resources.fx_queue_rotated.clear()
        self.camera = Vec2(-1.0, -1.0)
        if self.render_resources.ground is not None:
            terrain_seed = int(self.sim_world.state.rng.state)
            self.terrain_runtime.schedule_from_rng_seed(seed=terrain_seed, layers=3)

    def open_runtime(self) -> None:
        self.render_resources.texture_cache = self.texture_cache
        self.render_resources.config = self.config
        self.render_resources.open(terrain_seed=int(self.sim_world.state.rng.state))
        self.texture_cache = self.render_resources.texture_cache

    def close_runtime(self) -> None:
        self.render_resources.close()
        self.sim_world.close_session()

    def sync_audio_bridge_state(self) -> None:
        self.audio_bridge.sync(
            audio=self.audio,
            audio_rng=self.audio_rng,
            demo_mode_active=bool(self.demo_mode_active),
        )

    def update_camera(self, dt: float) -> None:
        _ = dt
        if not self.sim_world.players:
            return

        screen_size = self.renderer._camera_screen_size()
        alive = [player for player in self.sim_world.players if player.health > 0.0]
        if alive:
            inv_alive = 1.0 / float(len(alive))
            focus = Vec2(
                sum(player.pos.x for player in alive) * inv_alive,
                sum(player.pos.y for player in alive) * inv_alive,
            )
            camera = screen_size * 0.5 - focus
        else:
            camera = self.camera

        camera = camera + self.sim_world.state.camera_shake_offset
        self.camera = self.renderer._clamp_camera(camera, screen_size)

    def build_render_frame(self) -> RenderFrame:
        return self.render_resources.build_render_frame(
            state=self.sim_world.state,
            players=self.sim_world.players,
            creatures=self.sim_world.creatures,
            camera=self.camera,
            demo_mode_active=bool(self.demo_mode_active),
            elapsed_ms=float(self.sim_world.presentation_elapsed_ms),
            bonus_anim_phase=float(self.sim_world.bonus_anim_phase),
            lan_player_rings_enabled=bool(self.lan_player_rings_enabled),
            lan_local_aim_indicators_only=bool(self.lan_local_aim_indicators_only),
            lan_local_player_slot_index=int(self.lan_local_player_slot_index),
            rtx_mode=self.rtx_mode,
        )

    # ------------------------------------------------------------------
    # Tick-stepping
    # ------------------------------------------------------------------

    def init_tick_runner(
        self,
        *,
        game_mode: GameMode,
        build_inputs: WorldTickInputBuilder,
        tick_rate: int = 60,
    ) -> None:
        self._game_mode = game_mode
        self._build_inputs = build_inputs
        self._tick_rate = max(1, int(tick_rate))
        self._session: DeterministicSession | None = None
        self._runner: TickRunner | None = None
        self._world_state: object | None = None
        self._player_count = 0
        self._clock = FixedStepClock(tick_rate=int(self._tick_rate))
        self._frame_index = 0
        self._next_tick_index = 0

    def reset_tick_runner(self) -> None:
        self._session = None
        self._runner = None
        self._world_state = None
        self._player_count = 0
        self._clock = FixedStepClock(tick_rate=int(self._tick_rate))
        self._frame_index = 0
        self._next_tick_index = 0

    def _ensure_runner(self) -> tuple[TickRunner, DeterministicSession]:
        world_state = self.sim_world.world_state
        player_count = len(self.sim_world.players)
        session = self._session
        runner = self._runner
        if (
            session is not None
            and runner is not None
            and self._world_state is world_state
            and int(self._player_count) == int(player_count)
        ):
            return runner, session

        detail_preset = 5
        gore_disabled = 0
        config = self.config
        if config is not None:
            detail_preset = config.detail_preset
            gore_disabled = config.gore_disabled

        session = DeterministicSession(
            world=world_state,
            world_size=float(self.world_size),
            damage_scale_by_type=self.sim_world.damage_scale_by_type,
            fx_queue=self.render_resources.fx_queue,
            fx_queue_rotated=self.render_resources.fx_queue_rotated,
            game_mode=self._game_mode,
            detail_preset=int(detail_preset),
            gore_disabled=int(gore_disabled),
            game_tune_started=bool(self.sim_world.game_tune_started),
            demo_mode_active=bool(self.demo_mode_active),
            auto_pick_perks=False,
            perk_progression_enabled=False,
            apply_world_dt_steps=True,
            clear_fx_queues_each_tick=False,
        )
        provider = LocalInputProvider(
            player_count=int(player_count),
            build_inputs=self._build_inputs,
        )
        runner = TickRunner(
            session=session,
            input_provider=provider,
            config=TickRunnerConfig(),
        )
        self._session = session
        self._runner = runner
        self._world_state = world_state
        self._player_count = int(player_count)
        self._clock = FixedStepClock(tick_rate=int(self._tick_rate))
        self._frame_index = 0
        self._next_tick_index = 0
        return runner, session

    def _apply_tick_batch(
        self,
        *,
        batch: TickBatchResult,
        session: DeterministicSession,
    ) -> int:
        outputs = apply_sim_metadata_batch(
            sim_world=self.sim_world,
            completed_results=batch.completed_results,
            game_tune_started=bool(session.game_tune_started),
        )
        reactions = {
            int(result.source_tick.tick_index): build_post_apply_reaction(tick_result=result)
            for result in batch.completed_results
        }
        apply_presentation_outputs(
            outputs=outputs,
            sync_audio_bridge_state=self.sync_audio_bridge_state,
            apply_audio_plan=lambda plan, should_apply_audio: self.audio_bridge.apply_plan(
                plan=plan,
                apply_audio=bool(should_apply_audio),
            ),
            update_camera=self.update_camera,
            on_output_applied=lambda output: apply_post_apply_reaction(
                reaction=reactions.get(int(output.tick_index), PostApplyReaction()),
                play_sfx=self.audio_bridge.router.play_sfx,
            ),
            apply_audio=True,
        )
        return len(outputs)

    def advance_tick_frame(self, dt: float) -> int:
        if not self.sim_world.players:
            return 0
        self.terrain_runtime.process_pending()
        runner, session = self._ensure_runner()
        session.demo_mode_active = bool(self.demo_mode_active)
        dt = float(dt)
        ticks_requested = int(self._clock.advance(dt))
        advance = advance_tick_runner_frame(
            runner=runner,
            start_tick=int(self._next_tick_index),
            frame_index=int(self._frame_index),
            ticks_requested=int(ticks_requested),
            dt_seconds=float(dt),
            tick_dt_seconds=float(self._clock.dt_tick),
            is_networked=False,
            is_replay=False,
            refund_clock=self._clock,
        )
        self._frame_index = int(advance.frame_index)
        self._next_tick_index = int(advance.next_tick_index)
        return self._apply_tick_batch(
            batch=advance.batch,
            session=session,
        )



from __future__ import annotations

from collections.abc import Callable

import msgspec

from ..perks import PerkId
from ..perks.helpers import perk_active
from ..sim.state_types import GameplayState, PlayerState, WeaponSlot
from ..weapon_usage import weapon_usage_slot_for_weapon_id
from ..weapons import WEAPON_BY_ID, Weapon, WeaponId


def weapon_entry(weapon_id: WeaponId) -> Weapon:
    return WEAPON_BY_ID[weapon_id]


class _WeaponAssignCtx(msgspec.Struct):
    player: PlayerState
    clip_size: int


_WeaponAssignClipModifier = Callable[[_WeaponAssignCtx], None]


def _weapon_assign_clip_ammo_maniac(ctx: _WeaponAssignCtx) -> None:
    if perk_active(ctx.player, PerkId.AMMO_MANIAC):
        ctx.clip_size += max(1, int(float(ctx.clip_size) * 0.25))


def _weapon_assign_clip_my_favourite_weapon(ctx: _WeaponAssignCtx) -> None:
    if perk_active(ctx.player, PerkId.MY_FAVOURITE_WEAPON):
        ctx.clip_size += 2


_WEAPON_ASSIGN_CLIP_MODIFIERS: tuple[_WeaponAssignClipModifier, ...] = (
    _weapon_assign_clip_ammo_maniac,
    _weapon_assign_clip_my_favourite_weapon,
)


def init_default_alt_weapon(player: PlayerState) -> None:
    """Initialize native reset-time alternate weapon slot state."""

    player.alt_weapon = WeaponSlot(
        weapon_id=WeaponId.PISTOL,
        clip_size=12,
        ammo=12.0,
        reload_active=False,
        reload_timer=0.0,
        reload_timer_max=1.2,
        shot_cooldown=0.0,
    )


def weapon_assign_player(player: PlayerState, weapon_id: WeaponId, *, state: GameplayState) -> None:
    """Assign weapon and reset per-weapon runtime state (ammo/cooldowns)."""

    weapon_id = WeaponId(weapon_id)
    if state.status is not None and not state.demo_mode_active:
        usage_slot = weapon_usage_slot_for_weapon_id(int(weapon_id))
        if usage_slot is not None:
            state.status.increment_weapon_usage_slot(usage_slot)

    weapon = weapon_entry(weapon_id)
    player.weapon.weapon_id = weapon_id

    clip_size = int(weapon.clip_size)
    clip_ctx = _WeaponAssignCtx(player=player, clip_size=max(0, clip_size))
    for modifier in _WEAPON_ASSIGN_CLIP_MODIFIERS:
        modifier(clip_ctx)
    player.weapon.clip_size = max(0, int(clip_ctx.clip_size))
    player.weapon.ammo = float(player.weapon.clip_size)
    player.weapon_reset_latch = 0
    player.weapon.reload_active = False
    player.weapon.reload_timer = 0.0
    player.weapon.reload_timer_max = 0.0
    player.weapon.shot_cooldown = 0.0
    player.aux_timer = 2.0

    if state is not None:
        state.sfx_queue.append(weapon.reload_sound)


def most_used_weapon_id_for_player(
    state: GameplayState,
    *,
    player_index: int,
    fallback_weapon_id: WeaponId,
) -> WeaponId:
    """Return a weapon id for the player's most-used weapon."""

    idx = int(player_index)
    if 0 <= idx < len(state.weapon_shots_fired):
        counts = state.weapon_shots_fired[idx]
        if counts:
            start = 1 if len(counts) > 1 else 0
            best = max(range(start, len(counts)), key=lambda i: int(counts[i]))
            if int(counts[best]) > 0:
                return WeaponId(best)
    return WeaponId(fallback_weapon_id)


def player_swap_alt_weapon(player: PlayerState) -> bool:
    """Swap primary and alternate weapon runtime blocks (Alternate Weapon perk)."""

    if player.alt_weapon is None:
        return False
    player.weapon, player.alt_weapon = player.alt_weapon, player.weapon
    return True


def player_start_reload(player: PlayerState, state: GameplayState) -> None:
    """Start or refresh a reload timer (`player_start_reload` @ 0x00413430)."""

    if player.weapon.reload_active and (
        perk_active(player, PerkId.AMMUNITION_WITHIN) or perk_active(player, PerkId.REGRESSION_BULLETS)
    ):
        return

    weapon = weapon_entry(player.weapon.weapon_id)
    reload_time = float(weapon.reload_time)

    if not player.weapon.reload_active:
        player.weapon.reload_active = True

    if perk_active(player, PerkId.FASTLOADER):
        reload_time *= 0.7
    if state.bonuses.weapon_power_up > 0.0:
        reload_time *= 0.6

    player.weapon.reload_timer = max(0.0, reload_time)
    player.weapon.reload_timer_max = player.weapon.reload_timer



from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol

from ..batch_apply import PresentationTickOutput, SimMetadataSink, apply_tick_to_sim
from ..clock import FixedStepClock
from ..hooks import TickResult


class PlaybackFrameDriver(Protocol):
    def step_tick(self, tick_index: int) -> TickResult: ...


@dataclass(frozen=True, slots=True)
class PlaybackFrameAdvance:
    outputs: tuple[PresentationTickOutput, ...]
    tick_results: tuple[TickResult, ...]
    frame_index: int
    next_tick_index: int
    ticks_requested: int
    ticks_completed: int


def advance_playback_frame(
    *,
    driver: PlaybackFrameDriver,
    sim_world: SimMetadataSink,
    clock: FixedStepClock,
    start_tick: int,
    frame_index: int,
    dt_seconds: float,
    max_ticks: int | None,
    tick_limit: int,
    game_tune_started: bool,
) -> PlaybackFrameAdvance:
    ticks_requested = int(clock.advance(float(dt_seconds)))
    if max_ticks is not None:
        ticks_requested = min(int(ticks_requested), max(0, int(max_ticks)))

    next_frame_index = int(frame_index) + 1
    next_tick_index = int(start_tick)

    outputs: list[PresentationTickOutput] = []
    tick_results: list[TickResult] = []

    while len(tick_results) < ticks_requested and next_tick_index < int(tick_limit):
        tick_result = driver.step_tick(int(next_tick_index))
        apply_tick_to_sim(
            sim_world=sim_world,
            step=tick_result.payload.step,
            game_tune_started=bool(game_tune_started),
        )
        outputs.append(PresentationTickOutput(
            tick_index=int(tick_result.source_tick.tick_index),
            dt_sim=float(tick_result.payload.step.dt_sim),
            presentation=tick_result.payload.step.presentation,
        ))
        tick_results.append(tick_result)
        next_tick_index += 1

    ticks_completed = len(tick_results)
    if ticks_completed < ticks_requested and next_tick_index >= int(tick_limit):
        clock.accum += float(ticks_requested - ticks_completed) * float(clock.dt_tick)

    return PlaybackFrameAdvance(
        outputs=tuple(outputs),
        tick_results=tuple(tick_results),
        frame_index=int(next_frame_index),
        next_tick_index=int(next_tick_index),
        ticks_requested=int(ticks_requested),
        ticks_completed=int(ticks_completed),
    )



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable
from typing import Any

import msgspec

from ...render.world import profile_hooks


class RenderTelemetryFrameSnapshot(msgspec.Struct, frozen=True):
    frame_index: int
    tick_index_before_update: int
    tick_index_after_update: int
    update_ms: float
    draw_ms: float
    frame_ms: float
    draw_calls_total: int
    draw_calls_by_api: dict[str, int]
    draw_calls_by_pass: dict[str, int]
    pass_ms: dict[str, float]


class _SessionSink:
    def __init__(self, session: RenderTelemetrySession) -> None:
        self._session = session

    def on_pass_duration(self, pass_name: str, duration_ms: float) -> None:
        self._session.record_pass_duration(pass_name, duration_ms)


class RenderTelemetrySession:
    _INTERCEPT_APIS: tuple[tuple[str, str], ...] = (
        ("draw_texture_pro", "draw_texture_pro"),
        ("draw_circle", "draw_circle"),
        ("draw_circle_lines", "draw_circle_lines"),
        ("draw_circle_sector", "draw_circle_sector"),
        ("draw_ring", "draw_ring"),
        ("draw_rectangle", "draw_rectangle"),
        ("draw_rectangle_lines", "draw_rectangle_lines"),
        ("draw_text", "draw_text"),
        ("rl_begin", "rl_begin"),
        ("begin_blend_mode", "begin_blend_mode"),
        ("end_blend_mode", "end_blend_mode"),
    )

    def __init__(self) -> None:
        self._active = False
        self._originals: dict[str, Callable[..., Any]] = {}
        self._frames: list[RenderTelemetryFrameSnapshot] = []
        self._current_frame_index: int | None = None
        self._current_tick_before_update: int | None = None
        self._draw_calls_total = 0
        self._draw_calls_by_api: defaultdict[str, int] = defaultdict(int)
        self._draw_calls_by_pass: defaultdict[str, int] = defaultdict(int)
        self._pass_ms: defaultdict[str, float] = defaultdict(float)
        self._sink = _SessionSink(self)
        self._prev_sink: profile_hooks.RenderProfileSink | None = None

    @property
    def frames(self) -> tuple[RenderTelemetryFrameSnapshot, ...]:
        return tuple(self._frames)

    def start(self) -> None:
        if self._active:
            return

        from grim.raylib_api import rl

        self._active = True
        profile_hooks.clear_pass_stack()
        self._prev_sink = profile_hooks.set_active_sink(self._sink)
        for fn_name, api_name in self._INTERCEPT_APIS:
            original = getattr(rl, fn_name, None)
            if original is None or not callable(original):
                continue
            self._originals[fn_name] = original

            def _wrapped(*args: Any, _orig: Callable[..., Any] = original, _api_name: str = api_name, **kwargs: Any) -> Any:
                self.record_draw_call(_api_name)
                return _orig(*args, **kwargs)

            setattr(rl, fn_name, _wrapped)

    def stop(self) -> None:
        if not self._active:
            return

        from grim.raylib_api import rl

        self._active = False
        for fn_name, original in self._originals.items():
            setattr(rl, fn_name, original)
        self._originals.clear()
        profile_hooks.clear_pass_stack()
        profile_hooks.set_active_sink(self._prev_sink)
        self._prev_sink = None

    def __enter__(self) -> RenderTelemetrySession:
        self.start()
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> None:
        self.stop()

    def begin_frame(self, *, frame_index: int, tick_index_before_update: int) -> None:
        if not self._active:
            return
        self._current_frame_index = int(frame_index)
        self._current_tick_before_update = int(tick_index_before_update)
        self._draw_calls_total = 0
        self._draw_calls_by_api.clear()
        self._draw_calls_by_pass.clear()
        self._pass_ms.clear()

    def end_frame(
        self,
        *,
        tick_index_after_update: int,
        update_ms: float,
        draw_ms: float,
        frame_ms: float,
    ) -> None:
        if self._current_frame_index is None or self._current_tick_before_update is None:
            return
        snapshot = RenderTelemetryFrameSnapshot(
            frame_index=int(self._current_frame_index),
            tick_index_before_update=int(self._current_tick_before_update),
            tick_index_after_update=int(tick_index_after_update),
            update_ms=float(update_ms),
            draw_ms=float(draw_ms),
            frame_ms=float(frame_ms),
            draw_calls_total=int(self._draw_calls_total),
            draw_calls_by_api={str(key): int(value) for key, value in sorted(self._draw_calls_by_api.items())},
            draw_calls_by_pass={str(key): int(value) for key, value in sorted(self._draw_calls_by_pass.items())},
            pass_ms={str(key): float(value) for key, value in sorted(self._pass_ms.items())},
        )
        self._frames.append(snapshot)
        self._current_frame_index = None
        self._current_tick_before_update = None

    def record_draw_call(self, api_name: str) -> None:
        if self._current_frame_index is None:
            return
        api = str(api_name)
        self._draw_calls_total += 1
        self._draw_calls_by_api[api] += 1
        pass_name = profile_hooks.current_pass_name() or "_unscoped"
        self._draw_calls_by_pass[str(pass_name)] += 1

    def record_pass_duration(self, pass_name: str, duration_ms: float) -> None:
        if self._current_frame_index is None:
            return
        self._pass_ms[str(pass_name)] += float(duration_ms)


__all__ = ["RenderTelemetryFrameSnapshot", "RenderTelemetrySession"]



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass

from .hooks import TickResult
from .sessions import QuestSpawnState


@dataclass(frozen=True, slots=True)
class QuestPresentationReaction:
    play_hit_sfx: bool = False
    play_completion_music: bool = False


@dataclass(frozen=True, slots=True)
class PostApplyReaction:
    sfx_keys: tuple[str, ...] = ()
    quest: QuestPresentationReaction | None = None


def build_post_apply_reaction(
    *,
    tick_result: TickResult,
    quest_state: QuestSpawnState | None = None,
) -> PostApplyReaction:
    if quest_state is None:
        return PostApplyReaction(
            sfx_keys=tuple(str(key) for key in tick_result.payload.step.post_apply_sfx_keys),
        )
    return PostApplyReaction(
        sfx_keys=tuple(str(key) for key in tick_result.payload.step.post_apply_sfx_keys),
        quest=QuestPresentationReaction(
            play_hit_sfx=bool(quest_state.play_hit_sfx),
            play_completion_music=bool(quest_state.play_completion_music),
        ),
    )


def apply_post_apply_reaction(
    *,
    reaction: PostApplyReaction,
    play_sfx: Callable[[str], None] | None,
    play_completion_music: Callable[[], None] | None = None,
) -> None:
    quest = reaction.quest
    if play_sfx is not None:
        for key in reaction.sfx_keys:
            play_sfx(str(key))
        if quest is not None and bool(quest.play_hit_sfx):
            play_sfx("sfx_questhit")

    if quest is not None and bool(quest.play_completion_music) and play_completion_music is not None:
        play_completion_music()



from __future__ import annotations

from grim.geom import Vec2
from grim.math import clamp
from grim.raylib_api import rl

from ...creatures.anim import creature_anim_select_frame
from ...creatures.spawn import CreatureFlags, CreatureTypeId
from ...sim.world_defs import CREATURE_ANIM
from .constants import _RAD_TO_DEG
from .context import WorldRenderCtx


def draw_creature_sprite(
    render_ctx: WorldRenderCtx,
    texture: rl.Texture,
    *,
    type_id: CreatureTypeId,
    flags: CreatureFlags,
    phase: float,
    mirror_long: bool | None = None,
    shadow_alpha: int | None = None,
    pos: Vec2,
    screen_pos: Vec2 | None = None,
    rotation_rad: float,
    scale: float,
    size_scale: float,
    tint: rl.Color,
    shadow: bool = False,
) -> None:
    info = CREATURE_ANIM.get(type_id)
    if info is None:
        return
    mirror_flag = info.mirror if mirror_long is None else mirror_long
    # Long-strip mirroring is handled by frame index selection, not texture flips.
    index, _, _ = creature_anim_select_frame(
        phase,
        base_frame=info.base,
        mirror_long=mirror_flag,
        flags=flags,
    )
    if index < 0:
        return

    if screen_pos is None:
        screen_pos = render_ctx.world_to_screen(pos)
    width = float(texture.width) / 8.0 * size_scale * scale
    height = float(texture.height) / 8.0 * size_scale * scale
    src_x = float((index % 8) * (texture.width // 8))
    src_y = float((index // 8) * (texture.height // 8))
    src = rl.Rectangle(src_x, src_y, float(texture.width) / 8.0, float(texture.height) / 8.0)

    rotation_deg = float(rotation_rad * _RAD_TO_DEG)

    if shadow:
        # In the original exe this is a "darken" blend pass gated by fx_detail_0
        # (creature_render_type). We approximate it with a black silhouette draw.
        # The observed pass is slightly bigger than the main sprite and offset
        # down-right by ~1px at default sizes.
        alpha = int(shadow_alpha) if shadow_alpha is not None else int(clamp(float(tint.a) * 0.4, 0.0, 255.0) + 0.5)
        shadow_tint = rl.Color(0, 0, 0, alpha)
        shadow_scale = 1.07
        shadow_w = width * shadow_scale
        shadow_h = height * shadow_scale
        offset = width * 0.035 - 0.7 * scale
        shadow_dst = rl.Rectangle(screen_pos.x + offset, screen_pos.y + offset, shadow_w, shadow_h)
        shadow_origin = rl.Vector2(shadow_w * 0.5, shadow_h * 0.5)
        rl.draw_texture_pro(texture, src, shadow_dst, shadow_origin, rotation_deg, shadow_tint)

    dst = rl.Rectangle(screen_pos.x, screen_pos.y, width, height)
    origin = rl.Vector2(width * 0.5, height * 0.5)
    rl.draw_texture_pro(texture, src, dst, origin, rotation_deg, tint)



from __future__ import annotations

from collections.abc import Callable
from pathlib import Path

import msgspec

from grim.assets import PaqTextureCache, TextureLoader
from grim.config import CrimsonConfig
from grim.geom import Vec2
from grim.raylib_api import rl
from grim.terrain_render import GroundRenderer

from ..creatures.anim import creature_corpse_frame_for_type
from ..creatures.runtime import CreaturePool
from ..effects import FxQueue, FxQueueRotated
from ..gameplay import GameplayState
from ..render.frame import RenderFrame
from ..render.rtx.mode import RtxRenderMode
from ..render.terrain_fx import FxQueueTextures, bake_fx_queues
from ..sim.state_types import PlayerState
from ..sim.world_defs import CREATURE_ASSET


class RenderResources(msgspec.Struct):
    assets_dir: Path
    world_size: float = 1024.0
    texture_cache: PaqTextureCache | None = None
    config: CrimsonConfig | None = None

    ground: GroundRenderer | None = None
    fx_queue: FxQueue = msgspec.field(default_factory=FxQueue)
    fx_queue_rotated: FxQueueRotated = msgspec.field(default_factory=FxQueueRotated)
    fx_textures: FxQueueTextures | None = None
    creature_textures: dict[str, rl.Texture] = msgspec.field(default_factory=dict)
    projs_texture: rl.Texture | None = None
    particles_texture: rl.Texture | None = None
    bullet_texture: rl.Texture | None = None
    bullet_trail_texture: rl.Texture | None = None
    arrow_texture: rl.Texture | None = None
    bonuses_texture: rl.Texture | None = None
    bodyset_texture: rl.Texture | None = None
    clock_table_texture: rl.Texture | None = None
    clock_pointer_texture: rl.Texture | None = None
    aim_texture: rl.Texture | None = None
    muzzle_flash_texture: rl.Texture | None = None
    wicons_texture: rl.Texture | None = None

    _texture_loader: TextureLoader | None = None

    def _ensure_texture_loader(self) -> TextureLoader:
        if self._texture_loader is not None:
            return self._texture_loader
        if self.texture_cache is not None:
            loader = TextureLoader(
                assets_root=self.assets_dir,
                cache=self.texture_cache,
            )
        else:
            loader = TextureLoader.from_assets_root(self.assets_dir)
            if loader.cache is not None:
                self.texture_cache = loader.cache
        self._texture_loader = loader
        return loader

    def load_texture(self, name: str, *, cache_path: str) -> rl.Texture | None:
        loader = self._ensure_texture_loader()
        return loader.get(name=name, paq_rel=cache_path)

    def sync_ground_settings(self) -> None:
        if self.ground is None:
            return
        if self.config is None:
            self.ground.texture_scale = 1.0
            self.ground.screen_width = None
            self.ground.screen_height = None
            return
        self.ground.texture_scale = self.config.texture_scale
        self.ground.screen_width = float(self.config.screen_width)
        self.ground.screen_height = float(self.config.screen_height)

    def set_ground_textures(
        self,
        *,
        base: rl.Texture | None,
        overlay: rl.Texture | None,
        detail: rl.Texture | None,
    ) -> None:
        if base is None:
            return
        if self.ground is None:
            self.ground = GroundRenderer(
                texture=base,
                overlay=overlay,
                overlay_detail=detail,
                width=int(self.world_size),
                height=int(self.world_size),
                texture_scale=1.0,
                screen_width=None,
                screen_height=None,
            )
        else:
            self.ground.texture = base
            self.ground.overlay = overlay
            self.ground.overlay_detail = detail
        self.sync_ground_settings()

    def schedule_ground_generation(self, *, seed: int, layers: int = 3) -> None:
        if self.ground is None:
            return
        self.ground.schedule_generate(seed=int(seed), layers=int(layers))

    def process_ground_pending(self) -> None:
        if self.ground is None:
            return
        self.sync_ground_settings()
        self.ground.process_pending()

    def open(self, *, terrain_seed: int) -> None:
        self.close()
        self.creature_textures.clear()

        base = self.load_texture(
            "ter_q1_base",
            cache_path="ter/ter_q1_base.jaz",
        )
        overlay = self.load_texture(
            "ter_q1_tex1",
            cache_path="ter/ter_q1_tex1.jaz",
        )
        detail = overlay or base
        self.set_ground_textures(base=base, overlay=overlay, detail=detail)
        self.schedule_ground_generation(seed=int(terrain_seed), layers=3)

        for asset in sorted(set(CREATURE_ASSET.values())):
            texture = self.load_texture(
                asset,
                cache_path=f"game/{asset}.jaz",
            )
            if texture is not None:
                self.creature_textures[asset] = texture

        self.projs_texture = self.load_texture("projs", cache_path="game/projs.jaz")
        self.particles_texture = self.load_texture("particles", cache_path="game/particles.jaz")
        self.bullet_texture = self.load_texture("bullet_i", cache_path="load/bullet16.tga")
        self.bullet_trail_texture = self.load_texture("bulletTrail", cache_path="load/bulletTrail.tga")
        self.arrow_texture = self.load_texture("arrow", cache_path="load/arrow.tga")
        self.bonuses_texture = self.load_texture("bonuses", cache_path="game/bonuses.jaz")
        self.wicons_texture = self.load_texture("ui_wicons", cache_path="ui/ui_wicons.jaz")
        self.bodyset_texture = self.load_texture("bodyset", cache_path="game/bodyset.jaz")
        self.clock_table_texture = self.load_texture("ui_clockTable", cache_path="ui/ui_clockTable.jaz")
        self.clock_pointer_texture = self.load_texture("ui_clockPointer", cache_path="ui/ui_clockPointer.jaz")
        self.aim_texture = self.load_texture("ui_aim", cache_path="ui/ui_aim.jaz")
        self.muzzle_flash_texture = self.load_texture("muzzleFlash", cache_path="game/muzzleFlash.jaz")

        if self.particles_texture is not None and self.bodyset_texture is not None:
            self.fx_textures = FxQueueTextures(
                particles=self.particles_texture,
                bodyset=self.bodyset_texture,
            )
        else:
            self.fx_textures = None

    def close(self) -> None:
        if self.ground is not None and self.ground.render_target is not None:
            rl.unload_render_texture(self.ground.render_target)
            self.ground.render_target = None
        self.ground = None

        self._texture_loader = None
        self.creature_textures.clear()
        self.projs_texture = None
        self.particles_texture = None
        self.bullet_texture = None
        self.bullet_trail_texture = None
        self.arrow_texture = None
        self.bonuses_texture = None
        self.wicons_texture = None
        self.bodyset_texture = None
        self.clock_table_texture = None
        self.clock_pointer_texture = None
        self.aim_texture = None
        self.muzzle_flash_texture = None
        self.fx_textures = None
        self.fx_queue.clear()
        self.fx_queue_rotated.clear()

    def bake_fx_queues(
        self,
        *,
        corpse_frame_for_type: Callable[[int], int] = creature_corpse_frame_for_type,
    ) -> None:
        if self.ground is None or self.fx_textures is None:
            return
        if not (self.fx_queue.count or self.fx_queue_rotated.count):
            return
        bake_fx_queues(
            self.ground,
            fx_queue=self.fx_queue,
            fx_queue_rotated=self.fx_queue_rotated,
            textures=self.fx_textures,
            corpse_frame_for_type=corpse_frame_for_type,
        )

    def build_render_frame(
        self,
        *,
        state: GameplayState,
        players: list[PlayerState],
        creatures: CreaturePool,
        camera: Vec2,
        demo_mode_active: bool,
        elapsed_ms: float,
        bonus_anim_phase: float,
        lan_player_rings_enabled: bool,
        lan_local_aim_indicators_only: bool,
        lan_local_player_slot_index: int,
        rtx_mode: RtxRenderMode,
    ) -> RenderFrame:
        return RenderFrame(
            assets_dir=self.assets_dir,
            world_size=float(self.world_size),
            demo_mode_active=bool(demo_mode_active),
            config=self.config,
            camera=camera,
            ground=self.ground,
            state=state,
            players=players,
            creatures=creatures,
            creature_textures=self.creature_textures,
            projs_texture=self.projs_texture,
            particles_texture=self.particles_texture,
            bullet_texture=self.bullet_texture,
            bullet_trail_texture=self.bullet_trail_texture,
            arrow_texture=self.arrow_texture,
            bonuses_texture=self.bonuses_texture,
            bodyset_texture=self.bodyset_texture,
            clock_table_texture=self.clock_table_texture,
            clock_pointer_texture=self.clock_pointer_texture,
            aim_texture=self.aim_texture,
            muzzle_flash_texture=self.muzzle_flash_texture,
            wicons_texture=self.wicons_texture,
            elapsed_ms=float(elapsed_ms),
            bonus_anim_phase=float(bonus_anim_phase),
            lan_player_rings_enabled=bool(lan_player_rings_enabled),
            lan_local_aim_indicators_only=bool(lan_local_aim_indicators_only),
            lan_local_player_slot_index=int(lan_local_player_slot_index),
            rtx_mode=rtx_mode,
        )



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING

import msgspec

from .input_providers import ResolvedTick
from .sessions import DeterministicSessionTick

if TYPE_CHECKING:
    from .input_providers import GameCommand


class TickResult(msgspec.Struct):
    source_tick: ResolvedTick
    payload: DeterministicSessionTick
    replay_tick_index: int | None = None
    lan_sync: LanTickSync | None = None


@dataclass(frozen=True, slots=True)
class LanFrameSample:
    frame_tick_index: int
    frame_inputs: tuple[list[float], ...]
    commands: tuple[GameCommand, ...] = ()


@dataclass(frozen=True, slots=True)
class LanTickSync:
    frame_tick_index: int
    frame_inputs: tuple[list[float], ...]


@dataclass(slots=True)
class LanSyncCallbacks:
    role: str
    take_frame_sample: Callable[[int], LanFrameSample | None]
    broadcast_tick_frame: Callable[[int, tuple[list[float], ...], tuple[GameCommand, ...]], None] | None = None



from __future__ import annotations

from collections.abc import Callable

import msgspec

from grim.geom import Vec2

from ...owner_ref import OwnerRef
from ...sim.state_types import GameplayState, PlayerState

ProjectileSpawnFn = Callable[..., int]
OwnerRefForPlayerFn = Callable[[int], OwnerRef]
OwnerRefForPlayerProjectilesFn = Callable[[GameplayState, int], OwnerRef]


class PlayerPerkTickCtx(msgspec.Struct):
    state: GameplayState
    player: PlayerState
    player_pos_before_move: Vec2
    players: list[PlayerState] | None
    dt: float
    stationary: bool
    owner_ref_for_player: OwnerRefForPlayerFn
    owner_ref_for_player_projectiles: OwnerRefForPlayerProjectilesFn
    projectile_spawn: ProjectileSpawnFn



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import TypeAlias

import msgspec

from ..creatures.spawn import advance_survival_spawn_stage, tick_rush_mode_spawns, tick_survival_wave_spawns
from ..effects import FxQueue, FxQueueRotated
from ..game_modes import GameMode
from ..gameplay import survival_update_weapon_handouts
from ..perks.selection import perk_selection_current_choices, perk_selection_pick
from ..quests.runtime import tick_quest_completion_transition
from ..quests.timeline import quest_spawn_table_empty, tick_quest_mode_spawns
from ..quests.types import SpawnEntry
from .input import PlayerInput
from .input_providers import GameCommand, PerkMenuOpenCommand, PerkPickCommand
from .step_pipeline import (
    DeterministicStepResult,
    StepPipelineOptions,
    run_deterministic_step,
    time_scale_reflex_boost_factor,
)
from .timing import FrameTiming, zero_gate_active_from_state
from .world_state import WorldState

# ---------------------------------------------------------------------------
# Tick result types
# ---------------------------------------------------------------------------

class DeterministicSessionTick(msgspec.Struct):
    step: DeterministicStepResult
    elapsed_ms: float
    rng_marks: dict[str, int]
    creature_count_world_step: int



# ---------------------------------------------------------------------------
# Mid-/post-step hook system (spawn strategies)
# ---------------------------------------------------------------------------

class MidStepContext(msgspec.Struct, frozen=True):
    """Context passed to mid-step spawn hooks during deterministic stepping."""

    world: WorldState
    rng_marks: dict[str, int]
    elapsed_before_ms: float
    dt_sim_ms: float
    dt_raw_ms: float
    world_size: float

MidStepHook: TypeAlias = Callable[[MidStepContext], None]

class PostStepContext(msgspec.Struct, frozen=True):
    """Context passed to post-step hooks during deterministic stepping."""

    world: WorldState
    rng_marks: dict[str, int]
    step_result: DeterministicStepResult

PostStepHook: TypeAlias = Callable[[PostStepContext], None]

@dataclass
class SurvivalSpawnState:
    stage: int = 0
    spawn_cooldown_ms: float = 0.0

@dataclass
class RushSpawnState:
    spawn_cooldown_ms: float = 0.0

@dataclass
class QuestSpawnState:
    spawn_entries: tuple[SpawnEntry, ...] = ()
    spawn_timeline_ms: float = 0.0
    no_creatures_timer_ms: float = 0.0
    completion_transition_ms: float = -1.0
    completed: bool = False
    play_hit_sfx: bool = False
    play_completion_music: bool = False

def survival_mid_step(ctx: MidStepContext, spawn: SurvivalSpawnState) -> None:
    state = ctx.world.state
    survival_update_weapon_handouts(
        state,
        ctx.world.players,
        survival_elapsed_ms=ctx.elapsed_before_ms,
    )

    player_level = ctx.world.players[0].level if ctx.world.players else 1
    stage, milestone_calls = advance_survival_spawn_stage(spawn.stage, player_level=int(player_level))
    spawn.stage = stage
    for call in milestone_calls:
        ctx.world.creatures.spawn_template(
            call.template_id,
            call.pos,
            float(call.heading),
            state.rng,
            rand=state.rng.rand,
        )
    ctx.rng_marks["after_stage_spawns"] = int(state.rng.state)

    player_xp = ctx.world.players[0].experience if ctx.world.players else 0
    cooldown, wave_spawns = tick_survival_wave_spawns(
        spawn.spawn_cooldown_ms,
        ctx.dt_sim_ms,
        state.rng,
        player_count=len(ctx.world.players),
        survival_elapsed_ms=ctx.elapsed_before_ms,
        player_experience=int(player_xp),
        terrain_width=int(ctx.world_size),
        terrain_height=int(ctx.world_size),
    )
    spawn.spawn_cooldown_ms = cooldown
    ctx.world.creatures.spawn_inits(wave_spawns)
    ctx.rng_marks["after_wave_spawns"] = int(state.rng.state)

def rush_mid_step(ctx: MidStepContext, spawn: RushSpawnState) -> None:
    state = ctx.world.state
    cooldown, spawns = tick_rush_mode_spawns(
        spawn.spawn_cooldown_ms,
        ctx.dt_raw_ms,
        state.rng,
        player_count=len(ctx.world.players),
        survival_elapsed_ms=int(ctx.elapsed_before_ms),
        terrain_width=float(ctx.world_size),
        terrain_height=float(ctx.world_size),
    )
    spawn.spawn_cooldown_ms = cooldown
    ctx.world.creatures.spawn_inits(spawns)
    ctx.rng_marks["after_rush_spawns"] = int(state.rng.state)

def quest_post_step(ctx: PostStepContext, spawn: QuestSpawnState) -> None:
    state = ctx.world.state
    dt_ms = float(ctx.step_result.timing.dt_ms_i32)
    creatures_none_active = not any(c.active for c in ctx.world.creatures.entries)

    entries, timeline_ms, creatures_none_active, no_creatures_timer_ms, spawns = tick_quest_mode_spawns(
        spawn.spawn_entries,
        quest_spawn_timeline_ms=spawn.spawn_timeline_ms,
        frame_dt_ms=dt_ms,
        terrain_width=ctx.world.spawn_env.terrain_width,
        creatures_none_active=creatures_none_active,
        no_creatures_timer_ms=spawn.no_creatures_timer_ms,
    )
    spawn.spawn_entries = entries
    spawn.spawn_timeline_ms = float(timeline_ms)
    spawn.no_creatures_timer_ms = float(no_creatures_timer_ms)
    spawn_table_empty_now = quest_spawn_table_empty(spawn.spawn_entries)

    if not bool(state.demo_mode_active) and creatures_none_active and spawn_table_empty_now:
        state.bonuses.reflex_boost = 0.0
        state.time_scale_active = False

    for call in spawns:
        ctx.world.creatures.spawn_template(
            call.template_id,
            call.pos,
            float(call.heading),
            state.rng,
            rand=state.rng.rand,
        )
    ctx.rng_marks["after_quest_spawns"] = int(state.rng.state)

    any_alive_after = any(player.health > 0.0 for player in ctx.world.players)
    if any_alive_after:
        completion_ms, completed, play_hit_sfx, play_completion_music = tick_quest_completion_transition(
            spawn.completion_transition_ms,
            frame_dt_ms=dt_ms,
            creatures_none_active=creatures_none_active,
            spawn_table_empty=spawn_table_empty_now,
        )
        spawn.completion_transition_ms = float(completion_ms)
        spawn.completed = bool(completed)
        spawn.play_hit_sfx = bool(play_hit_sfx)
        spawn.play_completion_music = bool(play_completion_music)
    else:
        spawn.completion_transition_ms = -1.0
        spawn.completed = False
        spawn.play_hit_sfx = False
        spawn.play_completion_music = False

def rush_input_transform(inputs: list[PlayerInput]) -> list[PlayerInput]:
    return [
        msgspec.structs.replace(inp, reload_pressed=False) if inp.reload_pressed else inp
        for inp in inputs
    ]

# ---------------------------------------------------------------------------
# Shared timing helper
# ---------------------------------------------------------------------------

def _session_timing(state: object, dt: float) -> FrameTiming:
    """Compute frame timing from world state. Used by all session types."""
    return FrameTiming.compute(
        dt,
        time_scale_active_entry=bool(state.time_scale_active),  # type: ignore[union-attr]
        time_scale_factor=time_scale_reflex_boost_factor(
            reflex_boost_timer=float(state.bonuses.reflex_boost),  # type: ignore[union-attr]
            time_scale_active=bool(state.time_scale_active),  # type: ignore[union-attr]
        ),
        zero_gate_active=zero_gate_active_from_state(
            demo_mode_active=bool(state.demo_mode_active),  # type: ignore[union-attr]
        ),
    )

# ---------------------------------------------------------------------------
# Unified deterministic session (replaces Survival/Rush/Tutorial/Typo/WorldTick)
# ---------------------------------------------------------------------------

class DeterministicSession(msgspec.Struct):
    # Core state
    world: WorldState
    world_size: float
    damage_scale_by_type: dict[int, float]
    fx_queue: FxQueue
    fx_queue_rotated: FxQueueRotated

    # Mode identity
    game_mode: GameMode
    perk_progression_enabled: bool

    # Sim config
    detail_preset: int = 5
    gore_disabled: int = 0
    game_tune_started: bool = False
    demo_mode_active: bool = False
    auto_pick_perks: bool = False
    apply_world_dt_steps: bool = True
    defer_camera_shake_update: bool = False
    clear_fx_queues_each_tick: bool = False
    finalize_post_render_lifecycle: bool = False
    elapsed_uses_raw_dt: bool = False

    # Mutable timing
    elapsed_ms: float = 0.0

    # Optional hooks (provided by modes / callers)
    mid_step_hook: MidStepHook | None = None
    post_step_hook: PostStepHook | None = None
    before_step_hook: Callable[[], None] | None = None
    input_transform: Callable[[list[PlayerInput]], list[PlayerInput]] | None = None

    def timing_for_dt(self, dt: float) -> FrameTiming:
        return _session_timing(self.world.state, dt)

    def step_tick(
        self,
        *,
        timing: FrameTiming,
        inputs: list[PlayerInput] | None,
        trace_rng: bool = False,
        commands: list[GameCommand] | None = None,
    ) -> DeterministicSessionTick:
        if self.before_step_hook is not None:
            self.before_step_hook()

        tick_inputs = inputs
        if tick_inputs is not None and self.input_transform is not None:
            tick_inputs = self.input_transform(tick_inputs)

        post_apply_sfx_keys: list[str] = []
        for cmd in (commands or ()):
            match cmd:
                case PerkPickCommand(choice_index=ci):
                    picked = perk_selection_pick(
                        self.world.state,
                        self.world.players,
                        self.world.state.perk_selection,
                        ci,
                        game_mode=self.game_mode,
                        player_count=len(self.world.players),
                        dt=float(timing.dt_sim),
                        creatures=self.world.creatures.entries,
                    )
                    if picked is not None:
                        # Eagerly regenerate choices so the RNG sequence matches
                        # the recording path (pick sets choices_dirty=True).
                        perk_selection_current_choices(
                            self.world.state,
                            self.world.players,
                            self.world.state.perk_selection,
                            game_mode=self.game_mode,
                            player_count=len(self.world.players),
                        )
                        post_apply_sfx_keys.append("sfx_ui_bonus")
                case PerkMenuOpenCommand():
                    perk_selection_current_choices(
                        self.world.state,
                        self.world.players,
                        self.world.state.perk_selection,
                        game_mode=self.game_mode,
                        player_count=len(self.world.players),
                    )
                case _:
                    raise RuntimeError(f"unhandled command type: {type(cmd).__name__}")

        state = self.world.state
        dt_sim_ms = float(timing.dt_sim_ms_i32)
        dt_raw_ms = float(timing.dt_ms_i32)
        elapsed_before_ms = self.elapsed_ms

        rng_marks: dict[str, int] = {}
        if commands:
            rng_marks["after_commands"] = int(state.rng.state)
        rng_marks["before_world_step"] = int(state.rng.state)

        hook: Callable[[], None] | None = None
        if self.mid_step_hook is not None:
            ctx = MidStepContext(
                world=self.world,
                rng_marks=rng_marks,
                elapsed_before_ms=elapsed_before_ms,
                dt_sim_ms=dt_sim_ms,
                dt_raw_ms=dt_raw_ms,
                world_size=self.world_size,
            )
            _mid = self.mid_step_hook
            hook = lambda: _mid(ctx)  # noqa: E731

        step = run_deterministic_step(
            world=self.world,
            timing=timing,
            options=StepPipelineOptions(
                world_size=self.world_size,
                damage_scale_by_type=self.damage_scale_by_type,
                detail_preset=self.detail_preset,
                gore_disabled=self.gore_disabled,
                auto_pick_perks=self.auto_pick_perks,
                game_mode=self.game_mode,
                demo_mode_active=self.demo_mode_active,
                perk_progression_enabled=self.perk_progression_enabled,
                game_tune_started=self.game_tune_started,
            ),
            apply_world_dt_steps=self.apply_world_dt_steps,
            inputs=tick_inputs,
            fx_queue=self.fx_queue,
            fx_queue_rotated=self.fx_queue_rotated,
            defer_camera_shake_update=self.defer_camera_shake_update,
            mid_step_hook=hook,
            rng_marks_out=rng_marks,
            trace_presentation_rng=trace_rng,
        )
        if post_apply_sfx_keys:
            step = msgspec.structs.replace(
                step,
                post_apply_sfx_keys=tuple(str(key) for key in post_apply_sfx_keys),
            )
        if step.presentation.trigger_game_tune:
            self.game_tune_started = True

        if self.post_step_hook is not None:
            self.post_step_hook(
                PostStepContext(
                    world=self.world,
                    rng_marks=rng_marks,
                    step_result=step,
                ),
            )

        if self.clear_fx_queues_each_tick:
            self.fx_queue.clear()
            self.fx_queue_rotated.clear()

        creature_count_world_step = sum(1 for c in self.world.creatures.entries if c.active)
        rng_marks["after_world_step"] = int(state.rng.state)
        rng_marks["after_camera_update"] = int(rng_marks.get("ws_after_camera_update", state.rng.state))

        if self.finalize_post_render_lifecycle:
            self.world.creatures.finalize_post_render_lifecycle()

        dt_elapsed = dt_raw_ms if self.elapsed_uses_raw_dt else dt_sim_ms
        self.elapsed_ms = elapsed_before_ms + dt_elapsed

        return DeterministicSessionTick(
            step=step,
            elapsed_ms=self.elapsed_ms,
            rng_marks=rng_marks,
            creature_count_world_step=creature_count_world_step,
        )



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, cast

from grim.fonts.small import SmallFontData, draw_small_text, measure_small_text_width
from grim.geom import Vec2
from grim.raylib_api import rl

from ..assets import _ensure_texture_cache
from ..high_scores_layout import weapons_db_right_detail_x_shift
from ..types import FrontendContext
from .databases_base import _DatabaseBaseView

if TYPE_CHECKING:
    from ...weapons import Weapon


class UnlockedWeaponsDatabaseView(_DatabaseBaseView):
    def __init__(self, state: FrontendContext) -> None:
        super().__init__(state)
        self._wicons_tex: rl.Texture | None = None
        self._weapon_ids: list[int] = []
        self._selected_weapon_id: int | None = None
        self._list_scroll_index: int = 0

    def open(self) -> None:
        super().open()
        self._weapon_ids = self._build_weapon_database_ids()
        self._selected_weapon_id = None
        self._list_scroll_index = 0
        cache = _ensure_texture_cache(self.state)
        self._wicons_tex = cache.get_or_load("ui_wicons", "ui/ui_wicons.jaz").texture

    def close(self) -> None:
        self._wicons_tex = None
        self._selected_weapon_id = None
        super().close()

    def _back_button_pos(self) -> Vec2:
        # state_15: ui_buttonSm bbox [270,507]..[352,539] => relative to left panel (-98,194): (368, 313)
        return Vec2(368.0, 313.0)

    def _draw_contents(self, left_top_left: Vec2, right_top_left: Vec2, *, scale: float, font: SmallFontData) -> None:
        left = left_top_left
        right = right_top_left
        detail_shift_x = weapons_db_right_detail_x_shift(float(self.state.config.screen_width))
        detail_top_left = right + Vec2(detail_shift_x * scale, 0.0)
        text_scale = 1.0 * scale
        dim_color = rl.Color(255, 255, 255, int(255 * 0.7))
        text_color = rl.WHITE

        # state_15 title at (153,244) => relative to left panel (-98,194): (251,50)
        title_pos = left + Vec2(251.0 * scale, 50.0 * scale)
        title_text = "Unlocked Weapons Database"
        draw_small_text(
            font,
            title_text,
            title_pos,
            text_scale,
            rl.Color(255, 255, 255, 255),
        )
        title_w = measure_small_text_width(font, title_text, text_scale)
        # Decompile path draws a 1px outline strip under the title with alpha 0.5.
        rl.draw_rectangle_lines_ex(
            rl.Rectangle(
                title_pos.x,
                title_pos.y + 13.0 * scale,
                title_w,
                max(1.0, 1.0 * scale),
            ),
            1.0,
            rl.Color(255, 255, 255, int(255 * 0.5)),
        )

        weapon_ids = self._weapon_ids
        count = len(weapon_ids)
        weapon_label = "weapon" if count == 1 else "weapons"
        draw_small_text(
            font,
            f"{count} {weapon_label} in database",
            left + Vec2(210.0 * scale, 80.0 * scale),
            text_scale,
            dim_color,
        )
        draw_small_text(
            font,
            "Weapon",
            left + Vec2(210.0 * scale, 108.0 * scale),
            text_scale,
            text_color,
        )

        # Oracle frame: outer [114,322]-[364,486], inner [115,323]-[363,485].
        frame_x = left.x + 212.0 * scale
        frame_y = left.y + 128.0 * scale
        frame_w = 250.0 * scale
        frame_h = 164.0 * scale
        rl.draw_rectangle(int(round(frame_x)), int(round(frame_y)), int(round(frame_w)), int(round(frame_h)), rl.WHITE)
        rl.draw_rectangle(
            int(round(frame_x + 1.0 * scale)),
            int(round(frame_y + 1.0 * scale)),
            max(0, int(round(frame_w - 2.0 * scale))),
            max(0, int(round(frame_h - 2.0 * scale))),
            rl.BLACK,
        )

        # Oracle list widget is 10 rows tall.
        list_top_left = left + Vec2(218.0 * scale, 130.0 * scale)
        row_step = 16.0 * scale
        visible_rows = 10
        max_scroll = max(0, len(weapon_ids) - visible_rows)
        start = max(0, min(max_scroll, int(self._list_scroll_index)))
        end = min(len(weapon_ids), start + visible_rows)
        visible_weapon_ids = weapon_ids[start:end]
        for row, weapon_id in enumerate(visible_weapon_ids):
            name, _icon = self._weapon_label_and_icon(weapon_id)
            row_color = text_color if self._selected_weapon_id is not None and int(weapon_id) == int(self._selected_weapon_id) else dim_color
            draw_small_text(
                font,
                name,
                list_top_left.offset(dy=float(row) * row_step),
                text_scale,
                row_color,
            )

        if self._selected_weapon_id is None:
            return

        weapon_id = int(self._selected_weapon_id)
        name, icon_index = self._weapon_label_and_icon(weapon_id)
        weapon = self._weapon_entry(weapon_id)
        preserve_bugs = self.state.preserve_bugs
        weapon_no_label = "wepno" if preserve_bugs else "weapon"
        draw_small_text(
            font,
            f"{weapon_no_label} #{weapon_id}",
            detail_top_left + Vec2(240.0 * scale, 32.0 * scale),
            text_scale,
            rl.Color(255, 255, 255, int(255 * 0.4)),
        )
        draw_small_text(font, name, detail_top_left + Vec2(50.0 * scale, 50.0 * scale), text_scale, text_color)
        if icon_index is not None:
            self._draw_wicon(icon_index, pos=detail_top_left + Vec2(82.0 * scale, 82.0 * scale), scale=scale)

        reload_time = weapon.reload_time
        clip_size = weapon.clip_size
        ammo_class = int(weapon.ammo_class or 0)
        firerate_label = "Firerate" if preserve_bugs else "Fire rate"
        if ammo_class == 1:
            firerate_text = f"{firerate_label}: n/a"
        else:
            firerate_text = f"{firerate_label}: {self._weapon_rpm(weapon)} rpm"
        draw_small_text(
            font,
            firerate_text,
            detail_top_left + Vec2(66.0 * scale, 128.0 * scale),
            text_scale,
            text_color,
        )
        draw_small_text(
            font,
            f"Reload time: {reload_time:.1f} secs",
            detail_top_left + Vec2(66.0 * scale, 146.0 * scale),
            text_scale,
            text_color,
        )
        draw_small_text(
            font,
            f"Clip size: {clip_size}",
            detail_top_left + Vec2(66.0 * scale, 164.0 * scale),
            text_scale,
            text_color,
        )

    def _update_content_interaction(self, *, left_top_left: Vec2, scale: float, mouse: rl.Vector2) -> None:
        weapon_ids = self._weapon_ids
        if not weapon_ids:
            self._selected_weapon_id = None
            self._list_scroll_index = 0
            return

        visible_rows = 10
        max_scroll = max(0, len(weapon_ids) - visible_rows)
        mouse_wheel = int(rl.get_mouse_wheel_move())
        if mouse_wheel:
            self._list_scroll_index = max(0, min(max_scroll, int(self._list_scroll_index) - mouse_wheel))
        start = max(0, min(max_scroll, int(self._list_scroll_index)))
        end = min(len(weapon_ids), start + visible_rows)
        row_count = end - start
        if row_count <= 0:
            self._selected_weapon_id = None
            return

        row_step = 16.0 * scale
        list_hit_x = left_top_left.x + 214.0 * scale
        list_hit_y = left_top_left.y + 128.0 * scale
        list_hit_w = 246.0 * scale
        list_hit_h = min(160.0 * scale, row_step * float(row_count))
        if (
            list_hit_x <= mouse.x < list_hit_x + list_hit_w
            and list_hit_y <= mouse.y < list_hit_y + list_hit_h
        ):
            list_text_top = left_top_left.y + 130.0 * scale
            row = int((mouse.y - list_text_top) // row_step)
            if 0 <= row < row_count:
                self._selected_weapon_id = int(weapon_ids[start + row])
                return
        self._selected_weapon_id = None

    def _build_weapon_database_ids(self) -> list[int]:
        from ...weapon_usage import weapon_usage_slot_for_weapon_id
        from ...weapons import WEAPON_TABLE, WeaponId

        available: list[bool] | None = None
        from ...gameplay import WEAPON_COUNT_SIZE
        from ...weapon_runtime import (
            weapon_refresh_available as refresh_available,
        )

        weapon_refresh_available = cast(Callable[..., None], refresh_available)

        if weapon_refresh_available is not None:
            class _Stub:
                status: object | None
                game_mode: int
                demo_mode_active: bool
                weapon_available: list[bool]
                _weapon_available_game_mode: int
                _weapon_available_unlock_index: int
                _weapon_available_unlock_index_full: int

            stub = _Stub()
            stub.status = self.state.status
            stub.game_mode = self.state.config.game_mode
            stub.demo_mode_active = self.state.demo_enabled
            stub.weapon_available = [False] * int(WEAPON_COUNT_SIZE)
            stub._weapon_available_game_mode = -1
            stub._weapon_available_unlock_index = -1
            stub._weapon_available_unlock_index_full = -1
            try:
                weapon_refresh_available(stub)
                available = stub.weapon_available
            except (AttributeError, IndexError, TypeError, ValueError):
                available = None

        status = self.state.status
        used: list[int] = []
        for weapon in WEAPON_TABLE:
            weapon_id = int(weapon.weapon_id)
            include = False
            if available is not None:
                if 0 <= weapon_id < len(available):
                    include = bool(available[weapon_id])
            else:
                if weapon_id == WeaponId.PISTOL:
                    include = True
                else:
                    usage_slot = weapon_usage_slot_for_weapon_id(weapon_id)
                    include = bool(
                        status is not None
                        and usage_slot is not None
                        and status.weapon_usage_count_slot(usage_slot) != 0,
                    )
            if include:
                used.append(weapon_id)
        used.sort()
        return used

    def _weapon_entry(self, weapon_id: int) -> Weapon:
        from ...weapons import WEAPON_BY_ID, WeaponId

        return WEAPON_BY_ID[WeaponId(weapon_id)]

    def _weapon_rpm(self, weapon: Weapon) -> int:
        return int(60.0 / float(weapon.shot_cooldown))

    def _draw_wicon(self, icon_index: int, *, pos: Vec2, scale: float) -> None:
        tex = self._wicons_tex
        if tex is None:
            return
        idx = int(icon_index)
        if idx < 0 or idx > 31:
            return
        grid = 8
        cell_w = float(tex.width) / float(grid)
        cell_h = float(tex.height) / float(grid)
        frame = idx * 2
        src_x = float(frame % grid) * cell_w
        src_y = float(frame // grid) * cell_h
        icon_w = cell_w * 2.0
        icon_h = cell_h
        rl.draw_texture_pro(
            tex,
            rl.Rectangle(src_x, src_y, icon_w, icon_h),
            rl.Rectangle(pos.x, pos.y, icon_w * scale, icon_h * scale),
            rl.Vector2(0.0, 0.0),
            0.0,
            rl.WHITE,
        )

    def _weapon_label_and_icon(self, weapon_id: int) -> tuple[str, int | None]:
        from ...weapons import WEAPON_BY_ID, WeaponId, weapon_display_name

        weapon = WEAPON_BY_ID[WeaponId(weapon_id)]
        name = weapon_display_name(weapon.weapon_id, preserve_bugs=self.state.preserve_bugs)
        return name, weapon.icon_index



from __future__ import annotations

from pathlib import Path
from typing import Protocol

import msgspec


class ViewContext(msgspec.Struct, frozen=True):
    assets_dir: Path = Path("artifacts") / "assets"
    preserve_bugs: bool = False


class View(Protocol):
    def open(self) -> None: ...

    def update(self, dt: float) -> None: ...

    def draw(self) -> None: ...

    def close(self) -> None: ...



from __future__ import annotations

import tempfile
from pathlib import Path

from ..persistence.save_status import GameStatus
from ..status_snapshot import (
    game_status_from_progress_status,
    hash_progress_status,
    lockstep_status_from_progress,
    progress_status_from_game_status,
    progress_status_from_lockstep,
)
from .lockstep_protocol import StatusSnapshot


def status_snapshot_from_status(status: GameStatus | None) -> StatusSnapshot:
    return lockstep_status_from_progress(progress_status_from_game_status(status))


def hash_status_snapshot(snapshot: StatusSnapshot) -> str:
    """Stable hash for sanity-checking host/client status snapshots."""
    return hash_progress_status(progress_status_from_lockstep(snapshot))


def build_lan_deterministic_status(*, snapshot: StatusSnapshot | None = None) -> GameStatus:
    """Return a session-local `GameStatus` for deterministic LAN simulation.

    The host sends its save snapshot in `MatchStart`, and all peers use it as the
    simulation status. This avoids split brain where local save progress impacts
    deterministic simulation (weapon availability / RNG consumption).
    """

    # This status object should never overwrite the on-disk save. Give it a
    # safe temp path so accidental `save_if_dirty()` calls don't clobber
    # `game.cfg`.
    path = Path(tempfile.gettempdir()) / "crimson-lan-sim-game.cfg"
    progress = progress_status_from_lockstep(snapshot)
    return game_status_from_progress_status(progress, path=path)



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

import msgspec

from grim.fonts.small import SmallFontData
from grim.geom import Vec2

from ..frame import RenderFrame
from .context import WorldRenderCtx, build_world_render_ctx
from .draw import draw_world

if TYPE_CHECKING:
    from pathlib import Path

    from grim.config import CrimsonConfig

    from ...world.render_resources import RenderResources
    from ...world.sim_world_state import SimWorldState
    from ..rtx.mode import RtxRenderMode


class WorldRenderHost(Protocol):
    assets_dir: Path
    world_size: float
    demo_mode_active: bool
    config: CrimsonConfig | None
    camera: Vec2
    render_resources: RenderResources
    sim_world: SimWorldState
    lan_player_rings_enabled: bool
    lan_local_aim_indicators_only: bool
    lan_local_player_slot_index: int
    rtx_mode: RtxRenderMode

    def build_render_frame(self) -> RenderFrame: ...


class WorldRenderer(msgspec.Struct):
    _world: WorldRenderHost
    _render_frame: RenderFrame | None = None
    _small_font: SmallFontData | None = None

    def _active_render_ctx(self) -> WorldRenderCtx:
        return build_world_render_ctx(self, render_frame=self._render_frame)

    def draw(
        self,
        *,
        render_frame: RenderFrame | None = None,
        draw_aim_indicators: bool = True,
        entity_alpha: float = 1.0,
    ) -> None:
        frame = render_frame if render_frame is not None else self._world.build_render_frame()
        self._render_frame = frame
        try:
            render_ctx = build_world_render_ctx(self, render_frame=frame)
            draw_world(
                render_ctx,
                draw_aim_indicators=draw_aim_indicators,
                entity_alpha=entity_alpha,
            )
        finally:
            self._render_frame = None

    def _camera_screen_size(
        self,
        *,
        runtime_w: float | None = None,
        runtime_h: float | None = None,
    ) -> Vec2:
        return self._active_render_ctx()._camera_screen_size(runtime_w=runtime_w, runtime_h=runtime_h)

    def _clamp_camera(self, camera: Vec2, screen_size: Vec2) -> Vec2:
        return self._active_render_ctx()._clamp_camera(camera, screen_size)

    def _world_params(self) -> tuple[Vec2, Vec2]:
        return self._active_render_ctx()._world_params()

    def world_to_screen(self, pos: Vec2) -> Vec2:
        return self._active_render_ctx().world_to_screen(pos)

    def screen_to_world(self, pos: Vec2) -> Vec2:
        return self._active_render_ctx().screen_to_world(pos)



from __future__ import annotations

from grim.geom import Vec2
from grim.math import clamp
from grim.raylib_api import rl

from .constants import _RAD_TO_DEG
from .context import WorldRenderCtx


def grim2d_circle_segments_filled(radius: float) -> int:
    # grim_draw_circle_filled (grim.dll): segments = trunc(radius * 0.125 + 12.0)
    return max(3, int(radius * 0.125 + 12.0))


def grim2d_circle_segments_outline(radius: float) -> int:
    # grim_draw_circle_outline (grim.dll): segments = trunc(radius * 0.2 + 14.0)
    return max(3, int(radius * 0.2 + 14.0))


def draw_aim_circle(render_ctx: WorldRenderCtx, *, center: Vec2, radius: float, alpha: float = 1.0) -> None:
    if radius <= 1e-3:
        return
    alpha = clamp(float(alpha), 0.0, 1.0)
    if alpha <= 1e-3:
        return

    fill_a = int(77 * alpha + 0.5)  # ui_render_aim_indicators: rgba(0,0,0.1,0.3)
    outline_a = int(255 * 0.55 * alpha + 0.5)
    fill = rl.Color(0, 0, 26, fill_a)
    outline = rl.Color(255, 255, 255, outline_a)

    rl.begin_blend_mode(rl.BlendMode.BLEND_ALPHA)

    # The original uses a triangle fan (polygons). Raylib provides circle
    # primitives that still use triangles internally, but allow higher
    # segment counts for a smoother result when scaled.
    seg_count = max(grim2d_circle_segments_filled(radius), 64, int(radius))
    center_rl = center.to_rl()
    rl.draw_circle_sector(center_rl, float(radius), 0.0, 360.0, int(seg_count), fill)

    seg_count = max(grim2d_circle_segments_outline(radius), int(seg_count))
    # grim_draw_circle_outline draws a 2px-thick ring (outer radius = r + 2).
    # The exe binds bulletTrail, but that texture is white; the visual intent is
    # a subtle white outline around the filled spread circle.
    rl.draw_ring(center_rl, float(radius), float(radius + 2.0), 0.0, 360.0, int(seg_count), outline)

    rl.rl_set_texture(0)
    rl.end_blend_mode()


def draw_clock_gauge(
    render_ctx: WorldRenderCtx,
    *,
    pos: Vec2,
    ms: int,
    scale: float,
    alpha: float = 1.0,
) -> None:
    if render_ctx.clock_table_texture is None or render_ctx.clock_pointer_texture is None:
        return
    size = 32.0 * scale
    if size <= 1e-3:
        return
    tint = rl.Color(255, 255, 255, int(clamp(float(alpha), 0.0, 1.0) * 255.0 + 0.5))
    half = size * 0.5

    table_src = rl.Rectangle(
        0.0,
        0.0,
        float(render_ctx.clock_table_texture.width),
        float(render_ctx.clock_table_texture.height),
    )
    table_dst = rl.Rectangle(pos.x, pos.y, size, size)
    rl.draw_texture_pro(render_ctx.clock_table_texture, table_src, table_dst, rl.Vector2(0.0, 0.0), 0.0, tint)

    seconds = int(ms) // 1000
    pointer_src = rl.Rectangle(
        0.0,
        0.0,
        float(render_ctx.clock_pointer_texture.width),
        float(render_ctx.clock_pointer_texture.height),
    )
    pointer_dst = rl.Rectangle(pos.x + half, pos.y + half, size, size)
    origin = rl.Vector2(half, half)
    rotation_deg = float(seconds) * 6.0
    rl.draw_texture_pro(render_ctx.clock_pointer_texture, pointer_src, pointer_dst, origin, rotation_deg, tint)


def hud_indicator_enabled(render_ctx: WorldRenderCtx, player_index: int) -> bool:
    if render_ctx.config is None:
        return True
    raw = render_ctx.config.hud_indicators
    if not isinstance(raw, (bytes, bytearray)):
        return True
    idx = int(player_index)
    if idx < 0:
        return False
    if idx >= len(raw):
        return True
    return bool(raw[idx])


def direction_arrow_tint(render_ctx: WorldRenderCtx, player_index: int, *, alpha: float) -> rl.Color:
    alpha = clamp(float(alpha), 0.0, 1.0)
    if len(render_ctx.players) == 2:
        if int(player_index) == 0:
            return rl.Color(204, 230, 255, int(153.0 * alpha + 0.5))
        return rl.Color(255, 230, 204, int(153.0 * alpha + 0.5))
    return rl.Color(255, 255, 255, int(77.0 * alpha + 0.5))


def draw_direction_arrows(
    render_ctx: WorldRenderCtx,
    *,
    camera: Vec2,
    view_scale: Vec2,
    scale: float,
    alpha: float = 1.0,
) -> None:
    alpha = clamp(float(alpha), 0.0, 1.0)
    if alpha <= 1e-3:
        return
    arrow = render_ctx.arrow_texture
    if arrow is None:
        return

    src = rl.Rectangle(0.0, 0.0, float(arrow.width), float(arrow.height))
    width = max(1.0, float(arrow.width) * scale)
    height = max(1.0, float(arrow.height) * scale)
    origin = rl.Vector2(width * 0.5, height * 0.5)

    for player in render_ctx.players:
        if float(player.health) <= 0.0:
            continue
        index = int(player.index)
        if not hud_indicator_enabled(render_ctx, index):
            continue

        heading = float(player.heading)
        marker_pos = player.pos + Vec2.from_heading(heading) * 60.0
        screen = render_ctx._world_to_screen_with(marker_pos, camera=camera, view_scale=view_scale)
        dst = rl.Rectangle(screen.x, screen.y, width, height)
        tint = direction_arrow_tint(render_ctx, index, alpha=alpha)
        rl.draw_texture_pro(arrow, src, dst, origin, float(heading * _RAD_TO_DEG), tint)
